import numpy as np
from opentps.gui.viewer.dataForViewer.vtkSimpleImageFilter import VtkSimpleImageFilter

from opentps.core import Event
from opentps.gui.viewer.dataForViewer.genericImageForViewer import GenericImageForViewer
from opentps.gui.viewer.dataForViewer.image3DForViewer import Image3DForViewer


class Dyn3DSeqForViewer(GenericImageForViewer):
    def __init__(self, dyn3DSeq):
        super().__init__(dyn3DSeq)

        if hasattr(self, '_image3DForViewerList'):
            return

        self.dataChangedSignal = Event() # Not implemented in data class but required by opentps_core

        dyn3DSeq = self.data
        self._selectedPosition = np.array(dyn3DSeq.dyn3DImageList[0].origin) + np.array(dyn3DSeq.dyn3DImageList[0].gridSize) * np.array(dyn3DSeq.dyn3DImageList[0].spacing) / 2.0

        # This creates all image3DForViewers within the image sequence with the side effect that all VTK output port are initialized.
        self._image3DForViewerList = self._getImg3DForViewerList(dyn3DSeq.dyn3DImageList)

        self._simpleFilter = VtkSimpleImageFilter()
        self._currentIndexIn3DSeq = 0
        self._updateVTKOutputPort()
        self._vtkOutputPort = self._simpleFilter.GetOutputPort()
        self._range = (np.min(self.data.dyn3DImageList[0].imageArray), np.max(self.data.dyn3DImageList[0].imageArray))

    def _getImg3DForViewerList(self, dyn3DSeqImgList):
        vtkImageList = []
        for image in dyn3DSeqImgList:
            vtkImageList.append(Image3DForViewer(image))

        return vtkImageList

    @property
    def currentIndexIn3DSeq(self):
        return self._currentIndexIn3DSeq

    @currentIndexIn3DSeq.setter
    def currentIndexIn3DSeq(self, ind):
        if self._currentIndexIn3DSeq == ind:
            return

        self._currentIndexIn3DSeq = ind
        self._updateVTKOutputPort()

    def _updateVTKOutputPort(self):
        self._simpleFilter.RemoveAllInputs()
        currImageForViewer = self._image3DForViewerList[self._currentIndexIn3DSeq]
        self._simpleFilter.SetInputConnection(currImageForViewer.vtkOutputPort)

    @property
    def vtkOutputPort(self):
        return self._vtkOutputPort





import typing
from math import isclose

from opentps.core import Event

from opentps.gui.viewer.dataForViewer.dataMultiton import DataMultiton
import opentps.gui.viewer.dataViewerComponents.imageViewerComponents.lookupTables as lookupTables


class GenericImageForViewer(DataMultiton):
    def __init__(self, image):
        super().__init__(image)

        if hasattr(self, '_wwlValue'):
            return

        self.wwlChangedSignal = Event(tuple)
        self.lookupTableChangedSignal = Event(object)
        self.selectedPositionChangedSignal = Event(tuple)
        self.rangeChangedSignal = Event(tuple)

        self._range = (-500, 500)
        self._wwlValue = (self._range[1]-self._range[0], (self._range[1]+self._range[0])/2.)
        self._lookupTableName = 'gray'
        self._opacity = 0.5
        self._selectedPosition = (0, 0, 0)
        self._vtkOutputPort = None

        self._updateLT()

    @property
    def selectedPosition(self) -> tuple:
        return self._selectedPosition

    @selectedPosition.setter
    def selectedPosition(self, position: typing.Sequence):
        self._selectedPosition = (position[0], position[1], position[2])
        self.selectedPositionChangedSignal.emit(self._selectedPosition)

    @property
    def wwlValue(self) -> tuple:
        return self._wwlValue

    @wwlValue.setter
    def wwlValue(self, wwl: typing.Sequence):
        if isclose(wwl[0], self._wwlValue[0], abs_tol=0.1) and isclose(wwl[1], self._wwlValue[1], abs_tol=0.1):
            return

        self._wwlValue = (wwl[0], wwl[1])
        self.range = (wwl[1]-wwl[0]/2., wwl[1]+wwl[0]/2.)
        self.wwlChangedSignal.emit(self._wwlValue)

    @property
    def lookupTable(self) :
        return self._lookupTable

    @property
    def lookupTableName(self) -> str:
        return self._lookupTableName

    @lookupTableName.setter
    def lookupTableName(self, lookupTableName:str):
        if self._lookupTableName == lookupTableName:
            return

        self._lookupTableName = lookupTableName
        self._updateLT()

    def _updateLT(self):
        if self._lookupTableName == 'gray':
            self._lookupTable = lookupTables.grayLT(self._range)
        else:
            self._lookupTable = lookupTables.fusionLT(self._range, self._opacity, self._lookupTableName)
        self.lookupTableChangedSignal.emit(self._lookupTable)

    @property
    def range(self) -> tuple:
        return self._range

    @range.setter
    def range(self, range: typing.Sequence):
        if isclose(range[0], self._range[0], abs_tol=0.1) and isclose(range[1], self._range[1], abs_tol=0.1):
            return

        self._range = (range[0], range[1])
        self.wwlValue = (range[1]-range[0], (range[1]+range[0])/2.)

        if not (self._lookupTable is None) and self._lookupTableName=='gray':
            self._lookupTable.SetRange(self._range[0], self._range[1])
        else:
            self._updateLT()

        self.rangeChangedSignal.emit(self._range)

    @property
    def opacity(self) -> float:
        return self._opacity

    @opacity.setter
    def opacity(self, opacity: float):
        self._opacity = opacity
        self._updateLT()

    @property
    def vtkOutputPort(self):
        raise NotImplementedError()



from typing import Optional

import vtkmodules.vtkRenderingCore as vtkRenderingCore
from vtkmodules import vtkImagingCore, vtkCommonMath
from vtkmodules.vtkCommonColor import vtkNamedColors
from vtkmodules.vtkIOGeometry import vtkSTLReader
from vtkmodules.vtkInteractionWidgets import vtkOrientationMarkerWidget
from vtkmodules.vtkRenderingCore import vtkActor, vtkDataSetMapper

from opentps.core import Event
from opentps.gui.viewer.dataForViewer.genericImageForViewer import GenericImageForViewer
from opentps.gui.viewer.dataForViewer.image3DForViewer import Image3DForViewer


class PrimaryImage3DLayer:
    def __init__(self, renderer, renderWindow, iStyle):
        self.imageChangedSignal = Event(object)

        colors = vtkNamedColors()

        self._image = None
        self._iStyle = iStyle
        self._mainActor = vtkRenderingCore.vtkImageActor()
        self._mainMapper = self._mainActor.GetMapper()
        self._orientationActor = vtkActor()
        self._orientationMapper = vtkDataSetMapper()
        self._orientationWidget = vtkOrientationMarkerWidget()
        self._renderer = renderer
        self._renderWindow = renderWindow
        self._reslice = vtkImagingCore.vtkImageReslice()
        self._stlReader = vtkSTLReader()
        self._viewMatrix = vtkCommonMath.vtkMatrix4x4()

        self._mainMapper.SetSliceAtFocalPoint(True)

        self._orientationActor.SetMapper(self._orientationMapper)
        self._orientationActor.GetProperty().SetColor(colors.GetColor3d("Silver"))
        self._orientationMapper.SetInputConnection(self._stlReader.GetOutputPort())
        self._orientationWidget.SetViewport(0.8, 0.0, 1.0, 0.2)
        self._orientationWidget.SetCurrentRenderer(self._renderer)
        self._orientationWidget.SetInteractor(self._renderWindow.GetInteractor())
        self._orientationWidget.SetOrientationMarker(self._orientationActor)

        self._reslice.SetOutputDimensionality(2)
        self._reslice.SetInterpolationModeToNearestNeighbor()

        self._setMainMapperInputConnection()

    def _setMainMapperInputConnection(self):
        self._mainMapper.SetInputConnection(self._reslice.GetOutputPort())

    def close(self):
        self._disconnectAll()

    @property
    def image(self) -> Optional[Image3DForViewer]:
        """
        Image displayed
        :type:Optional[Image3DForViewer]
        """
        if self._image is None:
            return None

        return self._image

    @image.setter
    def image(self, image:Optional[GenericImageForViewer]):
        self._setImage(image)

    def _setImage(self, image:Optional[GenericImageForViewer]):
        if image == self._image:
            return

        if not (isinstance(image, GenericImageForViewer) or (image is None)):
            return

        self._image = image

        self._disconnectAll()
        self._renderer.RemoveActor(self._mainActor)
        self._reslice.RemoveAllInputs()

        if not (self._image is None):
            self._reslice.SetInputConnection(self._image.vtkOutputPort)

            self._renderer.AddActor(self._mainActor)

            self._image.lookupTableName = 'gray'
            self._setLookupTable()

            self._connectAll()

        self.imageChangedSignal.emit(self._image)

        self._renderWindow.Render()

    def _setLookupTable(self):
        imageProperty = self._mainActor.GetProperty()
        imageProperty.SetLookupTable(self._image.lookupTable)
        imageProperty.UseLookupTableScalarRangeOn()

    def _updateLookupTable(self, lt):
        imageProperty = self._mainActor.GetProperty()
        imageProperty.SetLookupTable(self._image.lookupTable)

        self._renderWindow.Render()

    @property
    def resliceAxes(self):
        """
        Reslice axes
        """
        return self._reslice.GetResliceAxes()

    @resliceAxes.setter
    def resliceAxes(self, resliceAxes):
        self._reslice.SetResliceAxes(resliceAxes)
        self._orientationActor.PokeMatrix(resliceAxes)

    def _connectAll(self):
        self._image.dataChangedSignal.connect(self._render)
        self._image.lookupTableChangedSignal.connect(self._updateLookupTable)
        self._image.rangeChangedSignal.connect(self._render)

    def _disconnectAll(self):
        if self._image is None:
            return

        self._image.dataChangedSignal.disconnect(self._render)
        self._image.lookupTableChangedSignal.disconnect(self._updateLookupTable)
        self._image.rangeChangedSignal.connect(self._render)

    def _render(self, *args):
        self._renderWindow.Render()

    def resliceDataFromPhysicalPoint(self, point):
        imageData = self._reslice.GetInput(0)

        ind = [0, 0, 0]
        imageData.TransformPhysicalPointToContinuousIndex(point, ind)
        return imageData.GetScalarComponentAsFloat(int(ind[0]), int(ind[1]), int(ind[2]), 0)



import numpy as np
from vtkmodules.vtkIOImage import vtkImageImport

from opentps.core import Event
from opentps.gui.viewer.dataForViewer.genericImageForViewer import GenericImageForViewer
from opentps.gui.viewer.dataViewerComponents.imageViewerComponents import lookupTables


class ROIMaskForViewer(GenericImageForViewer):
    def __init__(self, roiContour):
        super().__init__(roiContour)

        if hasattr(self, '_dataImporter'):
            return

        self.visibleChangedSignal = Event(bool)

        self._dataImporter = vtkImageImport()
        self._visible = False
        self._vtkOutputPort = None

        self.colorChangedSignal.connect(self._updateLT)

        self._updateVtkOutputPort()

    def __setattr__(self, key, value):
        super().__setattr__(key, value)

    @property
    def visible(self) -> bool:
        return self._visible

    @visible.setter
    def visible(self, visible: bool):
        self._visible = visible
        self.visibleChangedSignal.emit(self._visible)

    def _updateLT(self, *args):
        self._lookupTable = lookupTables.uniqueColorLT(1., 0.8, [self.color[0]/255, self.color[1]/255, self.color[2]/255])
        self.lookupTableChangedSignal.emit(self._lookupTable)

    def _updateVtkOutputPort(self):
        if self._imageArray is None:
            return
        referenceShape = self.gridSize
        referenceOrigin = self.origin
        referenceSpacing = self.spacing

        maskData = self._imageArray
        maskData = np.swapaxes(maskData, 0, 2)
        num_array = np.array(np.ravel(maskData), dtype=np.float32)

        self._dataImporter.SetNumberOfScalarComponents(1)
        self._dataImporter.SetDataScalarTypeToFloat()

        self._dataImporter.SetDataExtent(0, referenceShape[0] - 1, 0, referenceShape[1] - 1, 0, referenceShape[2] - 1)
        self._dataImporter.SetWholeExtent(0, referenceShape[0] - 1, 0, referenceShape[1] - 1, 0, referenceShape[2] - 1)
        self._dataImporter.SetDataSpacing(referenceSpacing[0], referenceSpacing[1], referenceSpacing[2])
        self._dataImporter.SetDataOrigin(referenceOrigin[0], referenceOrigin[1], referenceOrigin[2])

        data_string = num_array.tobytes()
        self._dataImporter.CopyImportVoidPointer(data_string, len(data_string))

        self._vtkOutputPort = self._dataImporter.GetOutputPort()

    @property
    def vtkOutputPort(self):
        return self._vtkOutputPort



from PyQt5.QtWidgets import QWidget,QVBoxLayout,QPushButton, QGridLayout,QLabel,QDoubleSpinBox, QSpinBox, QComboBox
import pyqtgraph as pg

from opentps.core.processing.deformableDataAugmentationToolBox.BreathingSignalGeneration import signalGeneration
from PyQt5.QtCore import Qt
class BreathingSignalPanel(QWidget):
    
    def __init__(self, viewController):
        QWidget.__init__(self)
        
        self._viewController = viewController
        
        self.layout = QVBoxLayout()
        self.layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(self.layout)
        
        self.Breath_param = {"Period": "", "Amplitude": "", "Duration": "", 
                             "Step": "", "Variance": "", "Mean": "", "ROI": "",
                             "nbrImage": "", "Way": ""}
        self.default_param = {"Period": 3.5, "Amplitude": 10.0, "Duration": 120, 
                             "Step": 0.2, "Variance": 0.0, "Mean": 0, 
                             "nbrImage": 3}

        self.layout.addWidget(QLabel('<b>Period:</b>'))
        self.TimePeriod = QDoubleSpinBox()
        self.TimePeriod.setGroupSeparatorShown(True)
        self.TimePeriod.setRange(0.0, 10.0)
        self.TimePeriod.setSingleStep(0.1)
        self.TimePeriod.setValue(self.default_param["Period"])
        self.TimePeriod.setSuffix(" seconds")
        self.TimePeriod.valueChanged.connect(self.Update_Breath_param)
        self.layout.addWidget(self.TimePeriod)
        self.layout.addSpacing(15)
        
        self.layout.addWidget(QLabel('<b>Amplitude:</b>'))
        self.Amplitude = QDoubleSpinBox()
        self.Amplitude.setGroupSeparatorShown(True)
        self.Amplitude.setRange(0.0, 30.0)
        self.Amplitude.setSingleStep(0.5)
        self.Amplitude.setValue(self.default_param["Amplitude"])
        self.Amplitude.setSuffix(" millimeter")
        self.Amplitude.valueChanged.connect(self.Update_Breath_param)
        self.layout.addWidget(self.Amplitude)
        self.layout.addSpacing(15)
        
        self.layout.addWidget(QLabel('<b>Duration:</b>'))
        self.Duration = QSpinBox()
        self.Duration.setGroupSeparatorShown(True)
        self.Duration.setRange(0, 600)
        self.Duration.setSingleStep(30)
        self.Duration.setValue(self.default_param["Duration"])
        self.Duration.setSuffix(" seconds")
        self.Duration.valueChanged.connect(self.Update_Breath_param)
        self.layout.addWidget(self.Duration)
        self.layout.addSpacing(15)
        
        self.layout.addWidget(QLabel('<b>Step size:</b>'))
        self.Step = QDoubleSpinBox()
        self.Step.setGroupSeparatorShown(True)
        self.Step.setRange(0.0, 1.0)
        self.Step.setSingleStep(0.05)
        self.Step.setValue(self.default_param["Step"])
        self.Step.valueChanged.connect(self.Update_Breath_param)
        self.layout.addWidget(self.Step)
        self.layout.addSpacing(15)
        
        self.layout.addWidget(QLabel('<b>Noise:</b>'))
        self.layout.addWidget(QLabel('Variance:'))
        self.Variance = QDoubleSpinBox()
        self.Variance.setGroupSeparatorShown(True)
        self.Variance.setRange(0.0, 10.0)
        self.Variance.setSingleStep(0.1)
        self.Variance.setValue(self.default_param["Variance"])
        self.Variance.valueChanged.connect(self.Update_Breath_param)
        self.layout.addWidget(self.Variance)
        self.layout.addWidget(QLabel('Mean:'))
        self.Mean = QSpinBox()
        self.Mean.setGroupSeparatorShown(True)
        self.Mean.setRange(0, 100)
        self.Mean.setSingleStep(1)
        self.Mean.setValue(self.default_param["Mean"])
        self.Mean.valueChanged.connect(self.Update_Breath_param)
        self.layout.addWidget(self.Mean)
        self.layout.addSpacing(15)
        
        self.layout.addWidget(QLabel('<b>ROI:</b>'))
        self.ROI = QComboBox()
        #self.ROI.setMaximumWidth(self.toolbox_width-18)
        self.ROI.currentIndexChanged.connect(self.Update_Breath_param)
        #self.ROI.addItems(["None"])
        self.layout.addWidget(self.ROI)
        self.layout.addSpacing(15)
        self.ROI_loaded = 0
        
        self.layout.addWidget(QLabel('<b>Number of images:</b>'))
        self.nbrImage = QSpinBox()
        self.nbrImage.setGroupSeparatorShown(True)
        self.nbrImage.setRange(0, 10)
        self.nbrImage.setSingleStep(1)
        self.nbrImage.setValue(self.default_param["nbrImage"])
        self.nbrImage.valueChanged.connect(self.Update_Breath_param)
        self.layout.addWidget(self.nbrImage)
        self.layout.addSpacing(15)
        
        self.layout.addWidget(QLabel('<b>Way:</b>'))
        self.way = QComboBox()
        #self.way.setMaximumWidth(self.toolbox_width-18)
        self.way.addItems(["Normal", "Random"])
        self.way.currentIndexChanged.connect(self.Update_Breath_param)
        self.layout.addWidget(self.way)
        self.layout.addSpacing(15)
        
        
        #self.Roi_name()#None if we do not load patient data
        #self.Update_Breath_param#in the case where we do not change the default values
        
        self.newSignalButton = QPushButton('Signal generation')
        self.layout.addWidget(self.newSignalButton)
        self.newSignalButton.clicked.connect(self.signalGeneration)
        
        self.layout.addStretch()
        
        self.xOnClick = 0
        self.yOnClick = 0

        
    def signalGeneration(self):       
        A = self.Amplitude.value() #amplitude (mm)
        dA = 5 #variation d amplitude possible (mm)
        T = self.TimePeriod.value() #periode respiratoire (s)
        df = 0.5 #variation de frequence possible (Hz)
        dS = 5 #shift du signal (mm)
        mean = self.Mean.value()
        sigma = self.Variance.value()
        step = self.Step.value() #periode d echantillonnage
        Tend = self.Duration.value() #temps de simulation
        L = 2/30 #moyenne des evenements aleatoires
        
        self.t,self.y = signalGeneration(A, dA, T, df, dS, mean, sigma, step, Tend, L)
        
        self.widget = QWidget()
        self.layout = QGridLayout()
        self.widget.setLayout(self.layout)
        self.coordSown = QLabel()
        self.graphWidget = pg.plot()#pg.GraphicsLayoutWidget()
        #self.graphWidget.showGrid(x = True, y = True)
        
        self.widget.setWindowTitle('Display the breathing signal')
        self.graphWidget.setBackground('w')
        self.widget.resize(1000, 1000) 
        cursor = Qt.CrossCursor
        self.graphWidget.setCursor(cursor)
        """
        self.plot_item = self.graphWidget.addPlot(title = "Breathing signal")
        self.plot_item.setLabel("left", 'Amplitude', units = 'mm')
        self.plot_item.setLabel("bottom", 'Time', units = 's')
        curve = self.plot_item.plot()
        curve.setData(self.t,self.y)

        """
        self.graphWidget.setLabel("left", 'Amplitude', units = 'mm')
        self.graphWidget.setLabel("bottom", 'Time', units = 's')
        self.graphWidget.setTitle("Breathing signal")
        self.graphWidget.plot(self.t,self.y)
        
        
        self.coordSown.setText("x={:.1f}".format(self.xOnClick) + ", " + "y={:.1f}".format(self.yOnClick))
        self.layout.addWidget(self.coordSown, 0, 0,1,2)
        self.layout.addWidget(self.graphWidget, 1, 0)
        
        self.graphWidget.scene().sigMouseClicked.connect(self.onClick)
        
        # buttons
        self.ButtonLayout = QGridLayout()
        
        self.ValidateButton = QPushButton('OK')
        self.ValidateButton.clicked.connect(self.acceptFunct)
        self.ButtonLayout.addWidget(self.ValidateButton, 1, 0)
        
        self.RefreshButton = QPushButton('Refresh')
        self.RefreshButton.clicked.connect(self.refreshFunct)
        self.ButtonLayout.addWidget(self.RefreshButton, 1, 1)
        
        self.CancelButton = QPushButton('Cancel')
        self.CancelButton.clicked.connect(self.rejectFunct)
        self.ButtonLayout.addWidget(self.CancelButton, 1, 2)
        
        self.layout.addLayout(self.ButtonLayout, 2, 0)
        self.Roi_name()#None if we do not load patient data
        #self.Update_Breath_param()#in the case where we do not change the default values
        self.widget.show()
        
    def Update_Breath_param(self):
        self.Breath_param["Period"] = self.TimePeriod.value()
        self.Breath_param["Amplitude"] = self.Amplitude.value()
        self.Breath_param["Duration"] = self.Duration.value()
        self.Breath_param["Step"] = self.Step.value()
        self.Breath_param["Variance"] = self.Variance.value()
        self.Breath_param["Mean"] = self.Mean.value()
        self.Breath_param["ROI"] = self.ROI.currentText()
        self.Breath_param["nbrImage"] = self.nbrImage.value()
        self.Breath_param["Way"] = self.way.currentText()
        #self.Compute_Breath(plot = False
        #self.Roi_name()
    def Roi_name(self):
        if self.ROI_loaded == 0:
            rts = self._viewController.currentPatient.rtStructs#.name
            listOfNames = []
            for i in range(len(rts)):
                for j in range(len(rts[i].contours)):
                    listOfNames.append(rts[i].contours[j].name)
                
            self.ROI.addItems(listOfNames)
            self.ROI_loaded = 1
    
    def acceptFunct(self):
        print("Chosen parameters ", self.Breath_param)
        self.widget.close()
    
    def rejectFunct(self):
        print("Reject breathing signal")
        self.widget.close()
        
    def refreshFunct(self):
        print("Refresh signal")
        self.signalGeneration()
        
    def onClick(self, event):
        #print("hello")
        p = self.graphWidget.plotItem.vb.mapSceneToView(event.scenePos())
        #p = self.graphWidget.vb.mapSceneToView(event.scenePos())
        #self.graphWidget.setText("x={:.1f}".format(p.x()))
        #self.label_y.setText("y={:.1f}".format(p.y()))
        self.coordSown.setText("x={:.1f}".format(p.x()) + ", " + "y={:.1f}".format(p.y()))
        #self.graphWidget.plotItem.setText("x={:.1f}".format(p.x()) + ", " + "y={:.1f}".format(p.y()))
        print(f"x: {p.x()}, y: {p.y()}")
        self.xOnClick = p.x()
        self.yOnClick = p.y()
        
        
        
    


from typing import Optional

from opentps.core.data.images import DoseImage
from opentps.core.data.images._image3D import Image3D
from opentps.gui.viewer.dataViewerComponents.doseComparisonImageProvider import DoseComparisonImageProvider
from opentps.gui.viewer.dataViewer import DataViewer


class DoseComparisonDataViewer(DataViewer):
    def __init__(self, viewController):
        self._displayDoseComparison = False
        self._doseComparisonImageProvider = DoseComparisonImageProvider()

        super().__init__(viewController)


    ####################################################################################################################
    # This is the logical part of the viewer. Should we migrate this to a dedicated controller?
    def _initializeControl(self):
        super()._initializeControl()

        self._imageViewerActions.doseComparisonDataViewer = self

        self.displayTypeChangedSignal.connect(self._handleDisplayTypeChange)

        self._viewController.dose1ChangedSignal.connect(self._setDose1)
        self._viewController.dose2ChangedSignal.connect(self._setDose2)

    @property
    def comparisonMetric(self):
        return self._doseComparisonImageProvider.comparisonMetric

    @comparisonMetric.setter
    def comparisonMetric(self, metric):
        self._doseComparisonImageProvider.comparisonMetric = metric

    def _handleDisplayTypeChange(self, displayType):
        super()._handleDisplayTypeChange(displayType)

        if not self._displayDoseComparison:
            self._doseComparisonImageProvider.doseComparisonImageChangedSignal.disconnect(self._handleNewDoseComparisonImage)

    def _handleNewDoseComparisonImage(self, *ags):
        self._setSecondaryImage(self._doseComparisonImageProvider.doseComparisonImage)

    def _setSecondaryImage(self, image:Optional[Image3D]):
        if not self._displayDoseComparison:
            super()._setSecondaryImage(image)

        if image != self._doseComparisonImageProvider.doseComparisonImage:
            self._displayDoseComparison = False

        if image is None:
            oldImage = self.cachedStaticImage3DViewer.secondaryImage
            if oldImage is None:
                return
        elif not (image.patient is None):
            image.patient.imageRemovedSignal.connect(self._removeImageFromViewers)

        self.cachedStaticImage3DViewer.secondaryImage = image

        if self._displayDoseComparison:
            self._imageViewerActions.setImageViewer(self._currentViewer)
            self._imageViewerActions.hide()
        else:
            self.cachedStaticDVHViewer.dose2 = None

    def _setDose1(self, image:Optional[DoseImage]):
        self._doseComparisonImageProvider.doseComparisonImageChangedSignal.disconnect(self._handleNewDoseComparisonImage)

        self._displayDoseComparison = True
        self._doseComparisonImageProvider.doseComparisonImageChangedSignal.connect(self._handleNewDoseComparisonImage)
        self._handleNewDoseComparisonImage()

        image.patient.imageRemovedSignal.connect(self._removeImageFromViewers)

        self._doseComparisonImageProvider.dose1 = image
        self._dvhViewer.dose = image

    def _setDose2(self, image:Optional[DoseImage]):
        self._doseComparisonImageProvider.doseComparisonImageChangedSignal.disconnect(self._handleNewDoseComparisonImage)

        self._displayDoseComparison = True
        self._doseComparisonImageProvider.doseComparisonImageChangedSignal.connect(self._handleNewDoseComparisonImage)
        self._handleNewDoseComparisonImage()

        image.patient.imageRemovedSignal.connect(self._removeImageFromViewers)

        self._doseComparisonImageProvider.dose2 = image
        self._dvhViewer.dose2 = image

    def _setDVHDose(self, image:Optional[DoseImage]):
        if not self._displayDoseComparison:
            self.cachedStaticDVHViewer.dose2 = None
            super()._setDVHDose(image)
        else:
            return



from enum import Enum

from PyQt5.QtWidgets import QWidget, QVBoxLayout

from opentps.gui.viewer.dataViewer import DroppedObject
from opentps.gui.viewer.gridFourElements import GridFourElements
from opentps.gui.viewer.viewerToolbar import ViewerToolbar

class ViewerPanel(QWidget):
    class LayoutTypes(Enum):
        DEFAULT = 'GRID_2BY2'
        GRID_2BY2 = 'GRID_2BY2'

    def __init__(self, viewController, parent):
        super().__init__(parent)

        self._layoutType = None
        self._viewerGrid = None

        self._viewController = viewController

        self._dropEnabled = False
        self._onDropEvent = None

        self._layout = QVBoxLayout(self)

        self._viewToolbar = ViewerToolbar(viewController, parent=self)
        self._layout.addWidget(self._viewToolbar)

        self._setLayoutType(self.LayoutTypes.DEFAULT)

        self._iniializeControl()

    def closeEvent(self, QCloseEvent):
        for element in self._viewerGrid.gridElements:
            element.close()
        super().closeEvent(QCloseEvent)

    def _dropEvent(self, e):
        if e.mimeData().hasText():
            if (e.mimeData().text() == DroppedObject.DropTypes.IMAGE):
                e.accept()
                if not self._onDropEvent is None:
                    self._onDropEvent(self._viewController.selectedImage)
                return
        e.ignore()

    @property
    def onDropEvent(self):
        """
            Function to execute on drop event
        """
        return self._onDropEvent

    @onDropEvent.setter
    def onDropEvent(self, func):
        self._onDropEvent = func

    @property
    def dropEnabled(self) -> bool:
        """
            Drag and drop enabled

            :type: bool
        """
        return self._dropEnabled

    @dropEnabled.setter
    def dropEnabled(self, enabled: bool):
        self._setDropEnabled(enabled)

    def _setDropEnabled(self, enabled: bool):
        self._dropEnabled = enabled

        if enabled:
            self._viewerGrid.setAcceptDrops(True)
            self._viewerGrid.dragEnterEvent = lambda event: event.accept()
            self._viewerGrid.dropEvent = lambda event: self._dropEvent(event)
        else:
            self._viewerGrid.setAcceptDrops(False)

    @property
    def layoutType(self)->LayoutTypes:
        """
            Layout type defined in ViewerPanel.LayoutTypes
        """
        return self._layoutType

    @layoutType.setter
    def layoutType(self, layoutType:LayoutTypes):
        self._setLayoutType(layoutType)

    def _setLayoutType(self, layoutType:LayoutTypes):
        self._layoutType = layoutType

        if not self._viewerGrid is None:
            self._layoutType.removeWidget(self._viewerGrid)

        if self._layoutType == self.LayoutTypes.GRID_2BY2:
            self._viewerGrid = GridFourElements(self._viewController, self)
            self._viewerGrid.setEqualSize()
        elif self._viewerGrid==None:
            return

        self._layout.addWidget(self._viewerGrid)


####################################################################################################################
    # This is the logical part of the viewer. Should we migrate this to a dedicated controller?
    def _iniializeControl(self):
        self._viewController.dynamicDisplayController.setToolBar(self._viewToolbar)
        self._viewController.independentViewsEnabledSignal.connect(lambda enabled: self._setDropEnabled(not enabled))
        self.onDropEvent = self._setViewControllerMainImage # It might seems weird to pass a function within the class but it is if someday we want to move the logical part out of this class.
        self._setDropEnabled(not self._viewController.independentViewsEnabled)

    def _setViewControllerMainImage(self, patientData):
        self._viewController.droppedImage = patientData




from PyQt5.QtWidgets import *
from PyQt5.QtCore import Qt
from opentps.core.data.plan import Robustness
from opentps.core.data.plan._robustnessProton import RobustnessProton
from opentps.core.data.plan._robustnessPhoton import RobustnessPhoton

class RobustnessSettings(QWidget):
    def __init__(self, viewController, planEvaluation=False, parent=None):
        QWidget.__init__(self, parent=parent)

        self.layout = QVBoxLayout(self)
        self.setLayout(self.layout)

        self.layout.setContentsMargins(0, 0, 0, 0)

        self.modality = "IMPT" #default
        self.planEval = planEvaluation

        self._robustnessSettingsButton = QPushButton('Modify robustness settings')
        self._robustnessSettingsButton.clicked.connect(self._openRobustnessSettings)
        self.layout.addWidget(self._robustnessSettingsButton)

        self._robustSettingsLabel = QLabel('')
        self.layout.addWidget(self._robustSettingsLabel)

        self._dialog = RobustnessSettingsDialog(self.modality, planEvaluation=self.planEval)
        self._robustParam = self._dialog.robustness
        RobustSettings = 'No robust settings'
        self._robustSettingsLabel.setText(RobustSettings)
        self._updateRobustSettings()

        #self._dialog = RobustnessSettingsDialog(modality="IMPT",planEvaluation=planEvaluation)
        #self._robustParam = self._dialog.robustness
        #self._robustParam = None
        #self._updateRobustSettings()
        
    @property
    def robustness(self) -> Robustness:
        return self._robustParam
    
    def _openRobustnessSettings(self):
        self._dialog = RobustnessSettingsDialog(self.modality, planEvaluation=self.planEval)
        #self._robustParam = self._dialog.robustness  # Always set this after creating the dialog
        if self._dialog.exec():
            self._robustParam = self._dialog.robustness
        self._updateRobustSettings()

    def _updateRobustSettings(self):
        RobustSettings = ''
        if (self._robustParam.selectionStrategy == self._robustParam.Strategies.ALL):
            if self.modality == "IMPT":
                self._robustParam.numScenarios = 81
                if(self._robustParam.setupSystematicError[0] == 0.0): self._robustParam.numScenarios/=3
                if(self._robustParam.setupSystematicError[1] == 0.0): self._robustParam.numScenarios/=3
                if(self._robustParam.setupSystematicError[2] == 0.0): self._robustParam.numScenarios/=3
                if(self._robustParam.rangeSystematicError == 0.0): self._robustParam.numScenarios/=3
            elif self.modality == "IMRT":
                self._robustParam.numScenarios = 27 if self._robustParam.setupSystematicError[0] != 0.0 else 1
            RobustSettings += '<b>Scenario</b>: '
            RobustSettings += 'All (' + str(int(self._robustParam.numScenarios)) + ' scenarios) <br>'
            RobustSettings += 'Syst. setup: E<sub>S</sub> = ' + str(self._robustParam.setupSystematicError) + ' mm<br>'
            if self.modality == "IMPT":
                RobustSettings += 'Rand. setup: &sigma;<sub>S</sub> = ' + str(self._robustParam.setupRandomError) + ' mm<br>'
                RobustSettings += 'Syst. range: E<sub>R</sub> = ' + str(self._robustParam.rangeSystematicError) + ' %<br>'
        elif (self._robustParam.selectionStrategy == self._robustParam.Strategies.REDUCED_SET):
            if self.modality == "IMPT":
                self._robustParam.numScenarios = 21
                if(self._robustParam.setupSystematicError[0] == 0.0): self._robustParam.numScenarios-=6
                if(self._robustParam.setupSystematicError[1] == 0.0): self._robustParam.numScenarios-=6
                if(self._robustParam.setupSystematicError[2] == 0.0): self._robustParam.numScenarios-=6
                if(self._robustParam.rangeSystematicError == 0.0): self._robustParam.numScenarios/=3
            elif self.modality == "IMRT":
                self._robustParam.numScenarios = 7 if self._robustParam.setupSystematicError[0] != 0.0 else 1
            RobustSettings += '<b>Scenario</b>: '
            RobustSettings += 'Reduced set (' + str(int(self._robustParam.numScenarios)) + ' scenarios) <br>'
            RobustSettings += 'Syst. setup: &Sigma;<sub>S</sub> = ' + str(self._robustParam.setupSystematicError) + ' mm<br>'
            if self.modality == "IMPT":
                RobustSettings += 'Rand. setup: &sigma;<sub>S</sub> = ' + str(self._robustParam.setupRandomError) + ' mm<br>'
                RobustSettings += 'Syst. range: &Sigma;<sub>R</sub> = ' + str(self._robustParam.rangeSystematicError) + ' %<br>'
        elif self._robustParam.selectionStrategy == self._robustParam.Strategies.RANDOM:
            RobustSettings += '<b>Scenario</b>: '
            RobustSettings += 'Random <br>'
            #RobustSettings += 'Syst. setup: &Sigma;<sub>S</sub> = ' + str(self._robustParam.setupSystematicError) + ' mm<br>'
            #RobustSettings += 'Rand. setup: &sigma;<sub>S</sub> = ' + str(self._robustParam.setupRandomError) + ' mm<br>'
            RobustSettings += 'Random num of scenarios = ' + str(int(self._robustParam.numRandomScenarios)) + '<br>'
            if self.modality == "IMPT": RobustSettings += 'Syst. range: &Sigma;<sub>R</sub> = ' + str(self._robustParam.rangeSystematicError) + ' %<br>'
        else:
            self._robustParam.selectionStrategy = self._robustParam.Strategies.DISABLED
            RobustSettings = 'No robust settings'

        self._robustSettingsLabel.setText(RobustSettings)

    def _updateForModality(self, modality):
        if modality == "IMRT":
            self.modality = "IMRT"
        elif modality == "IMPT":
            self.modality = "IMPT"
        else:
            raise ValueError("Unsupported modality selected")
    

class RobustnessSettingsDialog(QDialog):
    def __init__(self, modality, planEvaluation=False):
        super().__init__()
        self.modality = modality

        if self.modality == "IMRT":
            self._robustParam = RobustnessPhoton()
        elif self.modality == "IMPT":
            self._robustParam = RobustnessProton()
        else:
            raise ValueError("Unsupported modality selected")

        self._initialize_dialog()
        
    def _initialize_dialog(self):
        QDialog.__init__(self)
        self.setWindowTitle('Robustness Settings')
        self.resize(300, 300)
        self.main_layout = QVBoxLayout()
        self.setLayout(self.main_layout)

        self.main_layout.addWidget(QLabel('<b>Scenario selection strategy:</b>'))
        self._strategyBox = QComboBox()
        self._strategyBox.setMaximumWidth(300 - 18)

        self._strategyBox.addItems(['Disabled', 'Reduced set','All','Random'])

        self.main_layout.addWidget(self._strategyBox)
        self.main_layout.addSpacing(20)
        self.ErrorLayout = QGridLayout()
        self.main_layout.addLayout(self.ErrorLayout)
        self.ErrorLayout.addWidget(QLabel('<b>Setup errors:</b>'), 0, 0, 1, 2)
        self.ErrorLayout.addWidget(QLabel('X'), 0, 2, 1, 1, Qt.AlignCenter)
        self.ErrorLayout.addWidget(QLabel('Y'), 0, 3, 1, 1, Qt.AlignCenter)
        self.ErrorLayout.addWidget(QLabel('Z'), 0, 4, 1, 1, Qt.AlignCenter)
        self.ErrorLayout.setRowMinimumHeight(0, 25)
        self.ErrorLayout.addWidget(QLabel('Systematic'), 1, 0)
        self.SigmaS_label = QLabel('&Sigma;<sub>S</sub>')
        self.ErrorLayout.addWidget(self.SigmaS_label, 1, 1)
        self.syst_setup_x = QLineEdit(str(self._robustParam.setupSystematicError[0]))
        self.syst_setup_x.setMaximumWidth(30)
        self.ErrorLayout.addWidget(self.syst_setup_x, 1, 2)
        self.syst_setup_y = QLineEdit(str(self._robustParam.setupSystematicError[1]))
        self.syst_setup_y.setMaximumWidth(30)
        self.ErrorLayout.addWidget(self.syst_setup_y, 1, 3)
        self.syst_setup_z = QLineEdit(str(self._robustParam.setupSystematicError[2]))
        self.syst_setup_z.setMaximumWidth(30)
        self.ErrorLayout.addWidget(self.syst_setup_z, 1, 4)
        self.ErrorLayout.addWidget(QLabel('mm'), 1, 5)
        
        self.ErrorLayout.addWidget(QLabel('Equiv. margin:'), 3, 0, 1, 2)
        self.SetupMarginX = QLabel('1.0')
        self.ErrorLayout.addWidget(self.SetupMarginX, 3, 2)
        self.SetupMarginY = QLabel('1.0')
        self.ErrorLayout.addWidget(self.SetupMarginY, 3, 3)
        self.SetupMarginZ = QLabel('1.0')
        self.ErrorLayout.addWidget(self.SetupMarginZ, 3, 4)
        self.ErrorLayout.addWidget(QLabel('mm'), 3, 5)
        self.ErrorLayout.setRowMinimumHeight(3, 25)
        self.ErrorLayout.setRowMinimumHeight(4, 25)

        self.numRandSceLabel = QLabel('<b> Number of random scenarios: ')
        self.ErrorLayout.addWidget(self.numRandSceLabel)
        self.numRandSceSpin = QSpinBox()
        self.numRandSceSpin.setGroupSeparatorShown(True)
        self.numRandSceSpin.setRange(0, 100)
        self.numRandSceSpin.setSingleStep(1)
        self.numRandSceSpin.setValue(50)
        self.ErrorLayout.addWidget(self.numRandSceSpin)
        self.numRandSceLabel.hide()
        self.numRandSceSpin.hide()
        
        if self.modality == "IMPT":
            self.ErrorLayout.addWidget(QLabel('Random'), 2, 0)
            self.sigmaS_label = QLabel('&sigma;<sub>S</sub>')
            self.ErrorLayout.addWidget(self.sigmaS_label, 2, 1)
            self.rand_setup_x = QLineEdit(str(self._robustParam.setupRandomError[0]))
            self.rand_setup_x.setMaximumWidth(30)
            self.ErrorLayout.addWidget(self.rand_setup_x, 2, 2)
            self.rand_setup_y = QLineEdit(str(self._robustParam.setupRandomError[1]))
            self.rand_setup_y.setMaximumWidth(30)
            self.ErrorLayout.addWidget(self.rand_setup_y, 2, 3)
            self.rand_setup_z = QLineEdit(str(self._robustParam.setupRandomError[2]))
            self.rand_setup_z.setMaximumWidth(30)
            self.ErrorLayout.addWidget(self.rand_setup_z, 2, 4)
            self.ErrorLayout.addWidget(QLabel('mm'), 2, 5)

            self.ErrorLayout.addWidget(QLabel('<b>Range uncertainties:</b>'), 5, 0, 1, 4)
            self.ErrorLayout.addWidget(QLabel('Systematic'), 6, 0)
            self.SigmaR_label = QLabel('&Sigma;<sub>R</sub>')
            self.ErrorLayout.addWidget(self.SigmaR_label, 6, 1)
            self.syst_range = QLineEdit(str(self._robustParam.rangeSystematicError))
            self.syst_range.setMaximumWidth(30)
            self.ErrorLayout.addWidget(self.syst_range, 6, 2)
            self.ErrorLayout.addWidget(QLabel('%'), 6, 3)
            self.ErrorLayout.addWidget(QLabel('Equiv. error:'), 7, 0, 1, 2)
            self.RangeError = QLabel('1.0')
            self.ErrorLayout.addWidget(self.RangeError, 7, 2)
            self.ErrorLayout.addWidget(QLabel('%'), 7, 3)
            self.syst_range.textChanged.connect(self.recompute_margin)
        self.main_layout.addSpacing(30)

        self._strategyBox.currentIndexChanged.connect(self.updateRobustStrategy)
        self.syst_setup_x.textChanged.connect(self.recompute_margin)
        self.syst_setup_y.textChanged.connect(self.recompute_margin)
        self.syst_setup_z.textChanged.connect(self.recompute_margin)
        if self.modality=="IMPT":
            self.rand_setup_x.textChanged.connect(self.recompute_margin)
            self.rand_setup_y.textChanged.connect(self.recompute_margin)
            self.rand_setup_z.textChanged.connect(self.recompute_margin)
        self.recompute_margin()
        if (self._strategyBox.currentText() == 'Disabled'): self.updateRobustStrategy()

        # buttons
        self.ButtonLayout = QHBoxLayout()
        self.main_layout.addLayout(self.ButtonLayout)
        self.CancelButton = QPushButton('Cancel')
        self.ButtonLayout.addWidget(self.CancelButton)
        self.CancelButton.clicked.connect(self.reject)
        self.OkButton = QPushButton('OK')
        self.OkButton.clicked.connect(self.return_parameters)
        self.ButtonLayout.addWidget(self.OkButton)

        self.updateRobustStrategy()

    @property
    def robustness(self) -> Robustness:
        self._updateRobustParam()
        return self._robustParam

    @robustness.setter
    def robustness(self, r:Robustness):
        self._robustParam = r
        self.updateRobustStrategy()


    def _updateRobustParam(self):
        self._robustParam.setupSystematicError = [float(self.syst_setup_x.text()), float(self.syst_setup_y.text()),
                                                  float(self.syst_setup_z.text())]
        
        if self.modality=="IMPT":
            self._robustParam.setupRandomError = [float(self.rand_setup_x.text()), float(self.rand_setup_y.text()),
                                              float(self.rand_setup_z.text())]
            self._robustParam.rangeSystematicError = float(self.syst_range.text())
        self._robustParam.numRandomScenarios = self.numRandSceSpin.value()

        if (self._strategyBox.currentText() == 'Disabled'):
            self._robustParam.selectionStrategy = Robustness.Strategies.DISABLED
        elif (self._strategyBox.currentText() == 'All'):
            self._robustParam.selectionStrategy = Robustness.Strategies.ALL
        elif (self._strategyBox.currentText() == 'Reduced set'):
            self._robustParam.selectionStrategy = Robustness.Strategies.REDUCED_SET
        elif (self._strategyBox.currentText() == 'Random'):
            self._robustParam.selectionStrategy = Robustness.Strategies.RANDOM

    @property
    def robustStrategie(self):
        return self._robustParam.selectionStrategy

    @robustStrategie.setter
    def robustStrategie(self, strategy):
        if strategy == Robustness.Strategies.DISABLED:
            self._strategyBox.setCurrentText('Disabled')
        elif strategy == Robustness.Strategies.ALL:
            self._strategyBox.setCurrentText('ALL')
        elif strategy == Robustness.Strategies.REDUCED_SET:
            self._strategyBox.setCurrentText('Reduced set')
        elif strategy == Robustness.Strategies.RANDOM:
            self._strategyBox.setCurrentText('Random')

        self.updateRobustStrategy()

    def updateRobustStrategy(self):
        if (self._strategyBox.currentText() == 'Disabled'):
            self.syst_setup_x.setEnabled(False)
            self.syst_setup_y.setEnabled(False)
            self.syst_setup_z.setEnabled(False)

            self.numRandSceLabel.show()
            self.numRandSceSpin.show()
            self.numRandSceSpin.setEnabled(False)
            
            if self.modality=="IMPT":
                self.rand_setup_x.setEnabled(False)
                self.rand_setup_y.setEnabled(False)
                self.rand_setup_z.setEnabled(False)
                self.syst_range.setEnabled(False)

        elif (self._strategyBox.currentText() == 'All'):
            self.syst_setup_x.setEnabled(True)
            self.syst_setup_y.setEnabled(True)
            self.syst_setup_z.setEnabled(True)

            if self.modality=="IMPT":
                self.rand_setup_x.setEnabled(True)
                self.rand_setup_y.setEnabled(True)
                self.rand_setup_z.setEnabled(True)
                self.syst_range.setEnabled(True)
            self.SigmaS_label.setText('E<sub>S</sub>')
            
            self.syst_setup_x.setText('5.0')
            self.syst_setup_y.setText('5.0')
            self.syst_setup_z.setText('5.0')

            if self.modality=="IMPT": 
                self.rand_setup_x.setText('0.0')
                self.rand_setup_y.setText('0.0')
                self.rand_setup_z.setText('0.0')
                self.syst_range.setText('3.0')
                self.SigmaR_label.setText('E<sub>R</sub>')
            
            self.numRandSceLabel.hide()
            self.numRandSceSpin.hide()

        elif (self._strategyBox.currentText() == 'Reduced set'):
            self.syst_setup_x.setEnabled(True)
            self.syst_setup_y.setEnabled(True)
            self.syst_setup_z.setEnabled(True)
            if self.modality=="IMPT": 
                self.rand_setup_x.setEnabled(True)
                self.rand_setup_y.setEnabled(True)
                self.rand_setup_z.setEnabled(True)
                self.syst_range.setEnabled(True)
                self.SigmaR_label.setText('&Sigma;<sub>R</sub>')
            self.SigmaS_label.setText('&Sigma;<sub>S</sub>')
            self.syst_setup_x.setText('2.0')
            self.syst_setup_y.setText('2.0')
            self.syst_setup_z.setText('2.0')

            if self.modality=="IMPT": 
                self.rand_setup_x.setText('0.0')
                self.rand_setup_y.setText('0.0')
                self.rand_setup_z.setText('0.0')
                self.syst_range.setText('1.6')

            self.numRandSceLabel.hide()
            self.numRandSceSpin.hide()

        else:
            self.syst_setup_x.setEnabled(False)
            self.syst_setup_y.setEnabled(False)
            self.syst_setup_z.setEnabled(False)

            if self.modality=="IMPT":
                self.rand_setup_x.setEnabled(False)
                self.rand_setup_y.setEnabled(False)
                self.rand_setup_z.setEnabled(False) 
                self.syst_range.setEnabled(False)
                self.SigmaR_label.setText('&Sigma;<sub>R</sub>')
                self.rand_setup_x.setText('1.4')
                self.rand_setup_y.setText('1.4')
                self.rand_setup_z.setText('1.4')
            self.SigmaS_label.setText('&Sigma;<sub>S</sub>')  
            self.syst_setup_x.setText('1.6')
            self.syst_setup_y.setText('1.6')
            self.syst_setup_z.setText('1.6')
        

            self.numRandSceLabel.show()
            self.numRandSceSpin.show()
            self.numRandSceSpin.setEnabled(True)

            if self.modality=="IMPT": self.syst_range.setText('1.6')

        self.recompute_margin()

    def recompute_margin(self):
        Sigma_x = float(self.syst_setup_x.text())
        Sigma_y = float(self.syst_setup_y.text())
        Sigma_z = float(self.syst_setup_z.text())
        if self.modality=="IMPT":
            sigma_x = float(self.rand_setup_x.text())
            sigma_y = float(self.rand_setup_y.text())
            sigma_z = float(self.rand_setup_z.text())
        if hasattr(self,"syst_range"):
            range_sigma = float(self.syst_range.text())
        else:
            range_sigma = 1
        if not (hasattr(self,"sigma_x") and hasattr(self,"sigma_y") and hasattr(self,"sigma_z")):
            sigma_x=1
            sigma_y=1
            sigma_z=1

        if (self._strategyBox.currentText() == 'Reduced set'):
            margin_x = 1.0 * Sigma_x + 0.7 * sigma_x
            margin_y = 1.0 * Sigma_y + 0.7 * sigma_y
            margin_z = 1.0 * Sigma_z + 0.7 * sigma_z
            margin_r = 1.0 * range_sigma

        else:
            margin_x = 2.5 * Sigma_x + 0.7 * sigma_x
            margin_y = 2.5 * Sigma_y + 0.7 * sigma_y
            margin_z = 2.5 * Sigma_z + 0.7 * sigma_z
            margin_r = 1.5 * range_sigma

        if (self._strategyBox.currentText() != 'Disabled'):
            self.SetupMarginX.setText('{:3.1f}'.format(margin_x))
            self.SetupMarginY.setText('{:3.1f}'.format(margin_y))
            self.SetupMarginZ.setText('{:3.1f}'.format(margin_z))
            if self.modality=="IMPT": self.RangeError.setText('{:3.1f}'.format(margin_r))

    def return_parameters(self):
        self._updateRobustParam()
        self.accept()




from typing import Optional, Sequence

from vtkmodules import vtkImagingCore
from vtkmodules.vtkInteractionWidgets import vtkScalarBarWidget
from vtkmodules.vtkRenderingAnnotation import vtkScalarBarActor

from opentps.core import Event
from opentps.gui.viewer.dataForViewer.image2DForViewer import Image2DForViewer
from opentps.gui.viewer.dataViewerComponents.imageViewerComponents.primaryImage2DLayer import PrimaryImage2DLayer


class SecondaryImage2DLayer(PrimaryImage2DLayer):
    def __init__(self, renderer, renderWindow, iStyle):
        self._colorMapper = vtkImagingCore.vtkImageMapToColors()

        super().__init__(renderer, renderWindow, iStyle)

        self.colorbarVisibilitySignal = Event(bool)
        self.lookupTableChangedSignal = Event(bool)

        self._colorbarActor = vtkScalarBarActor()
        self._colorbarWidget = vtkScalarBarWidget()

        self._colorbarActor.SetNumberOfLabels(5)
        self._colorbarActor.SetOrientationToVertical()
        self._colorbarActor.SetVisibility(False)
        self._colorbarActor.SetUnconstrainedFontSize(14)
        self._colorbarActor.SetMaximumWidthInPixels(20)

        self._colorbarWidget.SetInteractor(self._renderWindow.GetInteractor())
        self._colorbarWidget.SetScalarBarActor(self._colorbarActor)

    def _setMainMapperInputConnection(self):
        # self._colorMapper.SetInputConnection(self._reslice.GetOutputPort())
        self._mainMapper.SetInputConnection(self._colorMapper.GetOutputPort())

    def close(self):
        super().close()

    def _setImage(self, image: Optional[Image2DForViewer]):
        if image == self._image:
            return

        super()._setImage(image)

        if image is None:
            self.colorbarOn = False
        else:
            self.colorbarOn = True # TODO: Get this from parent

        self._renderWindow.Render()

    @property
    def colorbarOn(self) -> bool:
        """
        Colorbar visibility
        :type: bool
        """
        return self._colorbarActor.GetVisibility()

    @colorbarOn.setter
    def colorbarOn(self, visible: bool):
        if visible==self._colorbarActor.GetVisibility():
            return

        if visible:
            self._colorbarActor.SetVisibility(True)
            self._colorbarWidget.On()
        else:
            self._colorbarActor.SetVisibility(False)
            self._colorbarWidget.Off()

        self.colorbarVisibilitySignal.emit(visible)

        self._renderWindow.Render()

    def _connectAll(self):
        super()._connectAll()

    def _disconnectAll(self):
        super()._disconnectAll()

        if self._image is None:
            return

    def _setLookupTable(self):
        self._image.lookupTableName = 'jet'
        self._colorMapper.SetLookupTable(self._image.lookupTable)
        self._colorbarActor.SetLookupTable(self._image.lookupTable)

    def _updateLookupTable(self, lt):
        self._colorMapper.SetLookupTable(lt)
        self._colorbarActor.SetLookupTable(lt)

        self._renderWindow.Render()

    def _setWWL(self, wwl: Sequence):
        # WWL is changed via iStyle. It is only working on the primary image.
        pass



from PyQt5.QtWidgets import QWidget


class Grid(QWidget):
    def __init__(self, viewController, parent):
        super().__init__(parent)

        self._gridElements = []
        self._viewController = viewController

    @property
    def gridElements(self):
        # Prevent from appending/removing to/from _gridElements
        return [element for element in self._gridElements]

    def appendGridElement(self, gridElement):
        self._gridElements.append(gridElement)

    def removeGridElement(self, gridElement):
        self._gridElements.remove(gridElement)



import copy
import os
import pickle
from typing import Sequence, Optional
import logging

from PyQt5.QtGui import QIcon
from PyQt5.QtWidgets import QTableWidget, QTableWidgetItem, QMainWindow, QAction, QFileDialog, QToolBar, QCheckBox, QLabel, QVBoxLayout, QWidget

from opentps.core.data.images import ROIMask
from opentps.core.data.plan import ProtonPlanDesign,Robustness
from opentps.core.data.plan import ObjectivesList
from opentps.core.processing.planOptimization.objectives.baseFunction import BaseFunc
from opentps.core.processing.planOptimization.objectives import dosimetricObjectives as dosObj
from opentps.core.data import Patient
from opentps.core import Event

import opentps.gui.res.icons as IconModule


class ObjectivesWindow(QMainWindow):
    def __init__(self, viewController, parent=None):
        super().__init__(parent)

        self.objectivesModifiedEvent = Event()

        self._viewController = viewController

        self._roitTable = ROITable(self._viewController, self)
        self._roitTable.objectivesModifiedEvent.connect(self.objectivesModifiedEvent.emit)
        self.setCentralWidget(self._roitTable._containerWidget)

        self._menuBar = QToolBar(self)
        self.addToolBar(self._menuBar)

        iconPath = IconModule.__path__[0] + os.path.sep

        self._openAction = QAction(QIcon(iconPath + 'folder-open.png'), "&Open template file", self)
        self._openAction.triggered.connect(self._handleOpen)
        self._menuBar.addAction(self._openAction)

        self._saveAction = QAction(QIcon(iconPath + 'disk.png'), "&Save template", self)
        self._saveAction.triggered.connect(self._handleSave)
        self._menuBar.addAction(self._saveAction)

    @property
    def patient(self):
        return self._roitTable.patient

    @patient.setter
    def patient(self, p:Patient):
        self._roitTable.patient = p

    @property
    def planDesign(self) -> ProtonPlanDesign:
        return self._roitTable.planDesign

    @planDesign.setter
    def planDesign(self, pd: ProtonPlanDesign):
        self._roitTable.planDesign = pd

    def getObjectiveTerms(self) -> Sequence[BaseFunc]:
        return self._roitTable.getObjectiveTerms()

    def _handleOpen(self):
        fname, _ = QFileDialog.getOpenFileName(self, 'Open file','c:\\', "Objective template")
        template = self._loadTemplate(fname)
        self._roitTable.applyTemplate(template)

    def _handleSave(self):
        fname, _ = QFileDialog.getSaveFileName(self, 'Save file','c:\\', "Objective template")
        self._saveTemplate(self._roitTable.getTemplate(), fname)

    def _saveTemplate(self, objectives:Sequence[BaseFunc], filePath:str):
        with open(filePath, 'wb') as f:
            pickle.dump(objectives, f)

    def _loadTemplate(self, filePath:str) -> Sequence[BaseFunc]:
        with open(filePath, 'rb') as f:
            res = pickle.load(f)

        return res


class ROITable(QTableWidget):
    DMIN_THRESH = 0.
    DMAX_THRESH = 999.
    DMEAN_THRESH = 999.
    D_DEFAULT = None
    DEFAULT_WEIGHT = 1.
    DEFAULT_PRESCRIPTION = None

    def __init__(self, viewController, parent=None):
        self._containerWidget = QWidget(parent)
        super().__init__(0, 10, parent)

        self.objectivesModifiedEvent = Event()
        self.robustnessEnabledEvent = Event(bool)

        self.setHorizontalHeaderLabels(['ROI', 'Robust', '*Target(s)', '**Prescription (Gy)', 'Weight', 'Dmin (Gy)', 'Weight', 'Dmax (Gy)', 'Weight', 'Dmean (Gy)'])
        self._roiCol = 0
        self._robustCol = 1
        self._TargetCol = 2
        self._PrescriptionCol = 3
        self._weightMinCol = 4
        self._dMinCol = 5
        self._weightMaxCol = 6
        self._dMaxCol = 7
        self._weightMeanCol = 8
        self._dMeanCol = 9

        self._planDesign = None
        self._patient:Optional[Patient] = None
        self._robustnessEnabled = True
        self._rois = []

        self._viewController = viewController

        self.cellChanged.connect(lambda *args: self.objectivesModifiedEvent.emit())

        self._infoLabel = QLabel("*The beamlets/spots will be placed in the selected targets. "
                            "**Prescriptions can only be added to targets and is MANDATORY.", self._containerWidget)
        layout = QVBoxLayout(self._containerWidget)
        layout.addWidget(self)
        layout.addWidget(self._infoLabel)
        self._containerWidget.setLayout(layout)

    def closeEvent(self, QCloseEvent):
        if not self._patient is None:
            self._patient.rtStructAddedSignal.disconnect(self.updateTable)
            self._patient.rtStructRemovedSignal.disconnect(self.updateTable)

        super().closeEvent(QCloseEvent)

    @property
    def planDesign(self) -> ProtonPlanDesign:
        return self._planDesign

    @planDesign.setter
    def planDesign(self, pd:ProtonPlanDesign):
        if self._planDesign is None:
            robustnessChanged = True
        else:
            robustnessChanged = self.robustnessEnabled != (self._planDesign.robustness.selectionStrategy != Robustness.Strategies.DISABLED)
        if self._planDesign==pd and (not robustnessChanged):
            return

        self.updateTable()
        self._planDesign = pd
        self.robustnessEnabled = self._planDesign.robustness.selectionStrategy != Robustness.Strategies.DISABLED
        self.applyTemplate(self._planDesign.objectives.objectivesList)

    @property
    def patient(self) -> Optional[Patient]:
        return self._patient

    @patient.setter
    def patient(self, p:Optional[Patient]):
        if p==self._patient:
            return

        if not self._patient is None:
            self._patient.rtStructAddedSignal.disconnect(self.updateTable)
            self._patient.rtStructRemovedSignal.disconnect(self.updateTable)

        self._patient = p

        if not self._patient is None:
            self._patient.rtStructAddedSignal.connect(self.updateTable)
            self._patient.rtStructRemovedSignal.connect(self.updateTable)

        self.updateTable()

    @property
    def robustnessEnabled(self):
        return self._robustnessEnabled

    @robustnessEnabled.setter
    def robustnessEnabled(self, enabled: bool):
        if self._robustnessEnabled == enabled:
            return

        for i, roi in enumerate(self._rois):
            if not enabled:
                self.cellWidget(i, self._robustCol).setChecked(False)
            self.cellWidget(i, self._robustCol).setEnabled(enabled)

        self._robustnessEnabled = enabled
        self.robustnessEnabledEvent.emit(self._robustnessEnabled)

    def updateTable(self, *args):
        self.reset()
        self._fillRoiTable()
        self.resizeColumnsToContents()
        self.resizeRowsToContents()
        self.objectivesModifiedEvent.emit()

    def _fillRoiTable(self):
        patient = self._viewController.currentPatient
        if patient is None:
            return

        rowCount = 0
        for rtStruct in patient.rtStructs:
            rowCount += len(rtStruct.contours)
        rowCount += len(patient.roiMasks)

        self.setRowCount(rowCount)

        self._rois = []
        i = 0
        for rtStruct in patient.rtStructs:
            for contour in rtStruct.contours:
                self.setItem(i, self._roiCol, QTableWidgetItem(contour.name))
                robustCheckBox = QCheckBox(self)
                robustCheckBox.setEnabled(self._robustnessEnabled)
                self.setCellWidget(i, self._robustCol, robustCheckBox)
                targetCheckBox = QCheckBox(self)
                targetCheckBox.setChecked(False)
                self.setCellWidget(i, self._TargetCol, targetCheckBox)
                prescriptionItem = QTableWidgetItem(str(self.DEFAULT_PRESCRIPTION))
                self.setItem(i, self._PrescriptionCol, prescriptionItem)
                self.setItem(i, self._weightMinCol, QTableWidgetItem(str(self.DEFAULT_WEIGHT)))
                self.setItem(i, self._dMinCol, QTableWidgetItem(str(self.D_DEFAULT)))
                self.setItem(i, self._weightMaxCol, QTableWidgetItem(str(self.DEFAULT_WEIGHT)))
                self.setItem(i, self._dMaxCol, QTableWidgetItem(str(self.D_DEFAULT)))
                self.setItem(i, self._weightMeanCol, QTableWidgetItem(str(self.DEFAULT_WEIGHT)))
                self.setItem(i, self._dMeanCol, QTableWidgetItem(str(self.D_DEFAULT)))

                self._rois.append(contour)

                i += 1

        for roiMask in patient.roiMasks:
            self.setItem(i, self._roiCol, QTableWidgetItem(roiMask.name))
            robustCheckBox = QCheckBox(self)
            robustCheckBox.setEnabled(self._robustnessEnabled)
            self.setCellWidget(i, self._robustCol, robustCheckBox)
            targetCheckBox = QCheckBox(self)
            targetCheckBox.setChecked(False)
            self.setCellWidget(i, self._TargetCol, targetCheckBox)
            self.setItem(i, self._PrescriptionCol, QTableWidgetItem(str(self.DEFAULT_PRESCRIPTION)))
            self.setItem(i, self._weightMinCol, QTableWidgetItem(str(self.DEFAULT_WEIGHT)))
            self.setItem(i, self._dMinCol, QTableWidgetItem(str(self.D_DEFAULT)))
            self.setItem(i, self._weightMaxCol, QTableWidgetItem(str(self.DEFAULT_WEIGHT)))
            self.setItem(i, self._dMaxCol, QTableWidgetItem(str(self.D_DEFAULT)))
            self.setItem(i, self._weightMeanCol, QTableWidgetItem(str(self.DEFAULT_WEIGHT)))
            self.setItem(i, self._dMeanCol, QTableWidgetItem(str(self.D_DEFAULT)))

            self._rois.append(roiMask)

            i += 1

    def applyTemplate(self, template:Sequence[BaseFunc]):
        roiNames = [roi.name for roi in self._rois]

        for obj in template:
            roiInd = roiNames.index(obj.roi.name)

            if obj.roi.name in roiNames:
                if obj.metric == "DMin":
                    self.item(roiInd, self._weightMinCol).setText(str(obj.weight))
                    self.item(roiInd, self._dMinCol).setText(str(obj.limitValue))
                elif obj.metric == "DMax":
                    self.item(roiInd, self._weightMaxCol).setText(str(obj.weight))
                    self.item(roiInd, self._dMaxCol).setText(str(obj.limitValue))
                elif obj.metric == "DMaxMean":
                    self.item(roiInd, self._weightMeanCol).setText(str(obj.weight))
                    self.item(roiInd, self._dMeanCol).setText(str(obj.limitValue))
                else:
                    pass
                    #TODO: metrics not supported
            else:
                self.setItem(roiInd, self._weightMinCol, QTableWidgetItem(str(self.DEFAULT_WEIGHT)))
                self.setItem(roiInd, self._dMinCol, QTableWidgetItem(str(self.DMIN_THRESH)))
                self.setItem(roiInd, self._weightMaxCol, QTableWidgetItem(str(self.DEFAULT_WEIGHT)))
                self.setItem(roiInd, self._dMaxCol, QTableWidgetItem(str(self.DMAX_THRESH)))
                self.setItem(roiInd, self._weightMeanCol, QTableWidgetItem(str(self.DEFAULT_WEIGHT)))
                self.setItem(roiInd, self._dMeanCol, QTableWidgetItem(str(self.DMEAN_THRESH)))

    def getTemplate(self) -> Sequence[BaseFunc]:
        objctivesToSave = []
        for objective in self.getObjectiveTerms():
            roi = objective.roi
            objective.roi = None
            objectiveCopy = copy.deepcopy(objective)
            objectiveCopy.roi = ROIMask(name=roi.name)
            objective.roi = roi
            objctivesToSave.append(objectiveCopy)

        return objctivesToSave

    def getObjectiveTerms(self) -> Sequence[BaseFunc]:
        terms = []

        for i, roi in enumerate(self._rois):
            # TODO How can this happen? It does happen when we load a new RTStruct for the same patient
            if self.cellWidget(i, self._TargetCol).isChecked():
                isTarget = True
                if self.item(i, self._PrescriptionCol).text() == 'None' :
                    logging.error('You need to give a float prescription for Target Volume.')
                    prescription = None
                else : 
                    prescription = float(self.item(i, self._PrescriptionCol).text())
            else:
                isTarget = False
                prescription = None

            robust = self.cellWidget(i, self._robustCol).isChecked()

            # Dmin
            if self.item(i, self._dMinCol).text() != 'None' :
                dmin = float(self.item(i, self._dMinCol).text())
                if dmin > self.DMIN_THRESH:
                    obj = dosObj.DMin(roi=roi,limitValue=dmin)
                    obj.weight = float(self.item(i, self._weightMinCol).text())
                    obj.robust = robust
                    obj.isTarget = isTarget
                    obj.prescription = prescription
                    terms.append(obj)
            # Dmax
            if self.item(i, self._dMaxCol).text() != 'None' :
                dmax = float(self.item(i, self._dMaxCol).text())
                if dmax < self.DMAX_THRESH:
                    obj = dosObj.DMax(roi=roi,limitValue= dmax)
                    obj.weight = float(self.item(i, self._weightMaxCol).text())
                    obj.robust = robust
                    obj.isTarget = isTarget
                    obj.prescription = prescription
                    terms.append(obj)
            # Dmean
            if self.item(i, self._dMeanCol).text() != 'None' :
                dmean = float(self.item(i, self._dMeanCol).text())
                if dmean < self.DMEAN_THRESH:
                    obj = dosObj.DMaxMean(roi=roi,limitValue=dmean)
                    obj.weight = float(self.item(i, self._weightMeanCol).text())
                    obj.robust = robust
                    obj.isTarget = isTarget
                    obj.prescription = prescription
                    terms.append(obj)

        return terms

    def getROIs(self):
        rois = []

        for i in range(len(self._rois)):
            # Dmin
            dmin = float(self.item(i, self._dMinCol).text())
            if dmin > self.DMIN_THRESH:
                rois.append(self._rois[i])
            # Dmax
            dmax = float(self.item(i, self._dMaxCol).text())
            if dmax < self.DMAX_THRESH:
                rois.append(self._rois[i])
            # Dmean
            dmean = float(self.item(i, self._dMeanCol).text())
            if dmean < self.DMEAN_THRESH:
                rois.append(self._rois[i])

        return rois

    def _setObjectives(self):
        objectiveList = ObjectivesList()
        for obj in self._objectivesWidget.objectives:
            objectiveList.append(obj)

        self._planDesign.objectives = objectiveList



from PyQt5.QtWidgets import *


class OptiSettingsDialog(QDialog):

    def __init__(self, optiParams):

        self.optiParam = optiParams

        # initialize the window
        QDialog.__init__(self)

        self.setWindowTitle('Optimization Settings')
        self.resize(300, 300)
        self.layout = QVBoxLayout()
        self.setLayout(self.layout)

        self.layout.addWidget(QLabel('<b>Optimization algorithm:</b>'))
        self.algorithm = QComboBox()
        self.algorithm.addItems(['Scipy-LBFGS (recommended)', 'Scipy-BFGS', 'In-house Gradient', 'In-house BFGS', 'In-house LBFGS', 'FISTA'])
        self.algorithm.setCurrentText(self.optiParam['method'])
        self.layout.addWidget(self.algorithm)

        self.layout.addSpacing(20)

        self._genParamLayout = QHBoxLayout()
        self.layout.addLayout(self._genParamLayout)

        self._genParamLayout.addWidget(QLabel('<b>General Parameters:</b>'))

        self._maxIterLayout = QHBoxLayout()
        self.layout.addLayout(self._maxIterLayout)

        self._maxIterLabel = QLabel('Max iterations:')
        self._maxIterLayout.addWidget(self._maxIterLabel)
        self._maxIterSpin = QDoubleSpinBox()
        self._maxIterSpin.setGroupSeparatorShown(True)
        self._maxIterSpin.setRange(0, 1e6)
        self._maxIterSpin.setSingleStep(50)
        self._maxIterSpin.setValue(1000)
        self._maxIterSpin.setDecimals(0)
        self._maxIterLayout.addWidget(self._maxIterSpin)

        self._stepLayout = QHBoxLayout()
        self.layout.addLayout(self._stepLayout)
        self._stepLabel = QLabel('Step:')
        self._stepLayout.addWidget(self._stepLabel)
        self._stepSpin = QDoubleSpinBox()
        self._stepSpin.setGroupSeparatorShown(True)
        self._stepSpin.setRange(0.0, 10.)
        self._stepSpin.setSingleStep(0.01)
        self._stepSpin.setValue(0.01)
        self._stepSpin.setEnabled(False)
        self._stepLayout.addWidget(self._stepSpin)
        self.layout.addSpacing(15)

        self._boundsLayout = QHBoxLayout()
        self.layout.addLayout(self._boundsLayout)
        self._boundsLabel = QLabel('MU Bounds constraints:')
        # self._boundsLabel.toggled.connect(self._setScoringSpacingVisible)
        self._boundsLayout.addWidget(self._boundsLabel)


        self._boundMinSpin = QDoubleSpinBox()
        self._boundMinSpin.setGroupSeparatorShown(True)
        self._boundMinSpin.setRange(0.0, 10.)
        self._boundMinSpin.setSingleStep(0.01)
        self._boundMinSpin.setValue(0.0)
        self._boundsLayout.addWidget(self._boundMinSpin)

        self._boundMaxSpin = QDoubleSpinBox()
        self._boundMaxSpin.setGroupSeparatorShown(True)
        self._boundMaxSpin.setRange(0.0, 1000.)
        self._boundMaxSpin.setSingleStep(5)
        self._boundMaxSpin.setValue(999.)
        self._boundsLayout.addWidget(self._boundMaxSpin)

        self.algorithm.setCurrentText(self.optiParam['method'])
        self.updateOptiParams()

        self.algorithm.currentIndexChanged.connect(self.updateOptiParams)

        # buttons
        self.buttonLayout = QHBoxLayout()
        self.layout.addLayout(self.buttonLayout)
        self.cancelButton = QPushButton('Cancel')
        self.buttonLayout.addWidget(self.cancelButton)
        self.cancelButton.clicked.connect(self.reject)
        self.okButton = QPushButton('OK')
        self.okButton.clicked.connect(self.returnParameters)
        self.buttonLayout.addWidget(self.okButton)

    def updateOptiParams(self):
        if self.algorithm.currentText() == 'Scipy-LBFGS (recommended)' or self.algorithm.currentText() == 'Scipy-BFGS':
            self._stepSpin.setEnabled(False)
            self._boundMinSpin.setEnabled(True)
            self._boundMaxSpin.setEnabled(True)
            self._boundsLabel.setEnabled(True)
        else:
            self._stepSpin.setEnabled(True)
            self._boundMinSpin.setEnabled(False)
            self._boundMaxSpin.setEnabled(False)
            self._boundsLabel.setEnabled(False)


    def returnParameters(self):

        self.optiParam["maxIter"] = int(self._maxIterSpin.value())
        try:
            self.optiParam["step"] = self._stepSpin.value()
        except ValueError:
            self.optiParam["step"] = None
        try:
            if self._boundMinSpin.value() == 0. and self._boundMaxSpin.value() == 999.:
                self.optiParam["bounds"] = None
            else:
                self.optiParam["bounds"] = self._boundMinSpin.value(), self._boundMaxSpin.value()
        except ValueError:
            self.optiParam["bounds"] = None
        print
        self.optiParam["method"] = self.algorithm.currentText()

        self.accept()




from opentps.gui import mainWindow, viewController
from opentps.gui.examples.extension.extension.gui.extensionPanel import ExtensionPanel


def addToGUI():
    mainWindow.mainToolbar.addWidget(ExtensionPanel(viewController), 'Extension example')


from PyQt5.QtWidgets import QComboBox
from opentps.core import Event
from opentps.core.data import Patient, PatientData


class PatientDataComboBox(QComboBox):
    def __init__(self, patientDataType, patient=None, parent=None):
        super().__init__(parent=parent)

        self.selectedDataEvent = Event()

        self._prevEventData = None

        self._patient = None
        self._patientDataType = patientDataType
        self._patientData = []

        self.currentIndexChanged.connect(self._handleNewData)
        self.currentTextChanged.connect(self._handleNewData)

        self.setPatient(patient)

    def closeEvent(self, QCloseEvent):
        self.setPatient(None)

        super().closeEvent(QCloseEvent)

    def _handleNewData(self):
        if self._prevEventData == self.selectedData:
            return

        self._prevEventData = self.selectedData
        self.selectedDataEvent.emit(self.selectedData)

    def setPatient(self, patient:Patient):
        if not (self._patient is None):
            self._patient.patientDataAddedSignal.disconnect(self._addData)
            self._patient.patientDataRemovedSignal.disconnect(self._removeData)

        self._patient = patient

        if self._patient is None:
            pass
        else:
            self._patient.patientDataAddedSignal.connect(self._addData)
            self._patient.patientDataRemovedSignal.connect(self._removeData)

            self._updateComboBox()

    @property
    def selectedData(self):
        if len(self._patientData)==0:
            return

        return self._patientData[self.currentIndex()]

    @selectedData.setter
    def selectedData(self, data):
        if len(self._patientData)==0:
            return

        try:
            currentIndex = self._patientData.index(data)
            self.setCurrentIndex(currentIndex)
        except:
            self.setCurrentIndex(0)
            if len(self._patientData):
                self.selectedData = self._patientData[0]

        self._handleNewData()

    def _updateComboBox(self):
        if self._checkIfSelfDeleted():
            return

        selectedData = self.selectedData

        self._removeAllData()

        for data in self._patient.getPatientDataOfType(self._patientDataType):
            self._addData(data)

        self.selectedData = selectedData


    def _addData(self, data:PatientData):
        if self._checkIfSelfDeleted():
            return

        if not(isinstance(data, self._patientDataType)):
            return

        self.addItem(data.name, data)
        self._patientData.append(data)
        data.nameChangedSignal.connect(self._handleDataChanged)

        self._handleNewData()

    def _removeData(self, data:PatientData):
        if self._checkIfSelfDeleted():
            return

        if not(isinstance(data, self._patientDataType)):
            return

        selectedData = self.selectedData

        data.nameChangedSignal.disconnect(self._handleDataChanged)
        self.removeItem(self.findData(data))
        self._patientData.remove(data)

        self.selectedData = selectedData

    def _removeAllData(self):
        for data in self._patientData:
            self._removeData(data)

    def _handleDataChanged(self, data):
        self._updateComboBox()

    def _checkIfSelfDeleted(self):
        try:
            self.findData(None)
        except:
            self.setPatient(None)
            return True

        return False



from PyQt5.QtWidgets import QWidget, QVBoxLayout, QComboBox, QLabel, QPushButton

from opentps.core.data.images import DoseImage
from opentps.core.data._patient import Patient
from opentps.gui.panels.patientDataWidgets import PatientDataComboBox


class DoseComparisonPanel(QWidget):
    def __init__(self, viewController):
        QWidget.__init__(self)

        self._patient:Patient = None
        self._viewController = viewController

        self.layout = QVBoxLayout()
        self.setLayout(self.layout)

        self._dose1Label = QLabel('Dose 1:')
        self.layout.addWidget(self._dose1Label)
        self._dose1ComboBox = PatientDataComboBox(DoseImage, patient=self._patient, parent=self)
        self.layout.addWidget(self._dose1ComboBox)

        self._dose2Label = QLabel('Dose 2:')
        self.layout.addWidget(self._dose2Label)
        self._dose2ComboBox = PatientDataComboBox(DoseImage, patient=self._patient, parent=self)
        self.layout.addWidget(self._dose2ComboBox)

        self._runButton = QPushButton('Update!')
        self._runButton.clicked.connect(self._run)
        self.layout.addWidget(self._runButton)

        self.layout.addStretch()

        self.setCurrentPatient(self._viewController.currentPatient)
        self._viewController.currentPatientChangedSignal.connect(self.setCurrentPatient)

    @property
    def _selectedDose1(self):
        return self._dose1ComboBox.selectedData

    @property
    def _selectedDose2(self):
        return self._dose2ComboBox.selectedData

    def setCurrentPatient(self, patient:Patient):
        self._patient = patient

        self._dose1ComboBox.setPatient(patient)
        self._dose2ComboBox.setPatient(patient)

    def _run(self):
        self._viewController.dose1 = self._selectedDose1
        self._viewController.dose2 = self._selectedDose2



import typing

from PyQt5.QtWidgets import *

import vtkmodules.vtkRenderingCore as vtkRenderingCore
import vtkmodules.vtkInteractionStyle as vtkInteractionStyle
from vtkmodules import vtkCommonMath
from vtkmodules.qt.QVTKRenderWindowInteractor import QVTKRenderWindowInteractor
from vtkmodules.vtkCommonCore import vtkCommand
from vtkmodules.vtkRenderingCore import vtkCoordinate

from opentps.core.data.images import Image2D
from opentps.core.data.plan._rtPlan import RTPlan
from opentps.core import Event
from opentps.gui.viewer.dataForViewer.genericImageForViewer import GenericImageForViewer
from opentps.gui.viewer.dataForViewer.image2DForViewer import Image2DForViewer
from opentps.gui.viewer.dataViewerComponents.imageViewerComponents.rtPlanLayer import RTPlanLayer
from opentps.gui.viewer.dataViewerComponents.blackEmptyPlot import BlackEmptyPlot
from opentps.gui.viewer.dataViewerComponents.imageViewerComponents.contourLayer import ContourLayer
from opentps.gui.viewer.dataViewerComponents.imageViewerComponents.crossHairLayer import CrossHairLayer
from opentps.gui.viewer.dataViewerComponents.imageViewerComponents.primaryImage2DLayer import PrimaryImage2DLayer
from opentps.gui.viewer.dataViewerComponents.profileWidget import ProfileWidget
from opentps.gui.viewer.dataViewerComponents.imageViewerComponents.secondaryImage2DLayer import SecondaryImage2DLayer
from opentps.gui.viewer.dataViewerComponents.imageViewerComponents.textLayer import TextLayer


class Image2DViewer(QWidget):
    # class ViewerTypes(Enum):
    #     AXIAL = 'axial'
    #     CORONAL = 'coronal'
    #     SAGITTAL = 'sagittal'
    #     DEFAULT = 'sagittal'
    #
    # _viewerTypesList = iter(list(ViewerTypes))


    def __init__(self, viewController):
        QWidget.__init__(self)

        self.crossHairEnabledSignal = Event(bool)
        self.profileWidgeEnabledSignal = Event(bool)
        self.wwlEnabledSignal = Event(bool)
        self.wwlEnabledSignal = Event(bool)
        self.viewTypeChangedSignal = Event(object)
        self.primaryImageSignal = Event(object)
        self.secondaryImageSignal = Event(object)

        self._blackWidget = BlackEmptyPlot(self)
        self._crossHairEnabled = False
        self._iStyle = vtkInteractionStyle.vtkInteractorStyleImage()
        self._leftButtonPress = False
        self._mainLayout = QVBoxLayout()
        self._profileWidgetNoInteractionYet = False # Used to know if initial position of profile widget must be set
        self._renderer = vtkRenderingCore.vtkRenderer()
        self.__sendingWWL = False
        self._viewController = viewController
        self._viewMatrix = vtkCommonMath.vtkMatrix4x4()
        # self._viewType = self.ViewerTypes.DEFAULT
        self._vtkWidget = QVTKRenderWindowInteractor(self)
        self._wwlEnabled = False

        self._renderWindow = self._vtkWidget.GetRenderWindow()

        self._crossHairLayer = CrossHairLayer(self._renderer, self._renderWindow)
        self._primaryImage2DLayer = PrimaryImage2DLayer(self._renderer, self._renderWindow, self._iStyle)
        self._secondaryImageLayer = SecondaryImage2DLayer(self._renderer, self._renderWindow, self._iStyle)
        self._profileWidget = ProfileWidget(self._renderer, self._renderWindow)
        self._textLayer = TextLayer(self._renderer, self._renderWindow)
        self._contourLayer = ContourLayer(self._renderer, self._renderWindow)
        self._rtPlanLayer = RTPlanLayer(self._renderer, self._renderWindow)

        self._profileWidget.primaryLayer = self._primaryImage2DLayer
        self._profileWidget.secondaryLayer = self._secondaryImageLayer
        self._profileWidget.contourLayer = self._contourLayer

        # self._setViewType(self._viewType)
        # self._contourLayer.resliceAxes = self._viewMatrix
        # self._rtPlanLayer.resliceAxes = self._viewMatrix

        self.setLayout(self._mainLayout)
        self._vtkWidget.hide()
        self._mainLayout.addWidget(self._blackWidget)
        self._blackWidget.show()
        self._mainLayout.setContentsMargins(0, 0, 0, 0)

        self._renderer.SetBackground(0, 0, 0)
        self._renderer.GetActiveCamera().SetParallelProjection(True)

        self._iStyle.SetInteractionModeToImageSlicing()

        self._iStyle.AddObserver("MouseWheelForwardEvent", self.onScroll)
        self._iStyle.AddObserver("MouseWheelBackwardEvent", self.onScroll)
        self._iStyle.AddObserver("MouseMoveEvent", self.onMouseMove)
        self._iStyle.AddObserver("LeftButtonPressEvent", self.onLeftButtonPressed)
        self._iStyle.AddObserver("LeftButtonReleaseEvent", self.onLeftButtonPressed)

        self._renderWindow.GetInteractor().SetInteractorStyle(self._iStyle)
        self._renderWindow.AddRenderer(self._renderer)

        self._closed = False

    def closeEvent(self, QCloseEvent):
        if self._closed:
            return

        self._closed = True
        self.reset()
        self._renderWindow.Finalize()
        self._vtkWidget.close()
        del self._renderWindow, self._vtkWidget
        super().closeEvent(QCloseEvent)

    def reset(self):
        if not (self._primaryImage2DLayer.image is None):
            self._primaryImage2DLayer.image.selectedPositionChangedSignal.disconnect(self._handlePosition)
            self._primaryImage2DLayer.image.nameChangedSignal.disconnect(self._setPrimaryName)

        if not (self._secondaryImageLayer.image is None):
            self._secondaryImageLayer.image.nameChangedSignal.disconnect(self._setSecondaryName)

        self._primaryImage2DLayer.close()
        self._secondaryImageLayer.close()
        self._textLayer.close()
        self._contourLayer.close()
        self._crossHairLayer.close()
        self._rtPlanLayer.close()

    @property
    def rtPlan(self) -> RTPlan:
        raise NotImplementedError

    @rtPlan.setter
    def rtPlan(self, plan:RTPlan):
        if not self.primaryImage is None:
            self._rtPlanLayer.setPlan(plan, self.primaryImage)

    @property
    def primaryImage(self) -> Image2D:
        if self._primaryImage2DLayer.image is None:
            return None
        return self._primaryImage2DLayer.image.data

    @primaryImage.setter
    def primaryImage(self, image: Image2D):
        imageAlreadyDisplayed = image == self._primaryImage2DLayer.image or (not (self._primaryImage2DLayer.image is None) and image == self._primaryImage2DLayer.image.data)
        if imageAlreadyDisplayed:
            return

        if image is None:
            self._resetPrimaryImageLayer()
        else:
            self._setPrimaryImageForViewer(Image2DForViewer(image))

        self.primaryImageSignal.emit(self.primaryImage)

    def _resetPrimaryImageLayer(self):
        self._primaryImage2DLayer.image = None
        self._mainLayout.removeWidget(self._vtkWidget)
        self._vtkWidget.hide()
        self._mainLayout.addWidget(self._blackWidget)
        self._blackWidget.show()

    def _setPrimaryImageForViewer(self, image:GenericImageForViewer):

        print('in image2DVIewer _setPrimaryImageForViewer', type(image))

        self._primaryImage2DLayer.image = image
        self._contourLayer.referenceImage = image
        self._textLayer.setPrimaryTextLine(2, image.name)

        self._handlePosition(self._primaryImage2DLayer.image.selectedPosition)

        if not (self._primaryImage2DLayer.image is None):
            self._primaryImage2DLayer.image.selectedPositionChangedSignal.disconnect(self._handlePosition)
            self._primaryImage2DLayer.image.nameChangedSignal.disconnect(self._setPrimaryName)

        self._primaryImage2DLayer.image.selectedPositionChangedSignal.connect(self._handlePosition)
        self._primaryImage2DLayer.image.nameChangedSignal.connect(self._setPrimaryName)

        # self._primaryImage2DLayer.resliceAxes = self._viewMatrix
        self._contourLayer.resliceAxes = self._viewMatrix
        self._rtPlanLayer.resliceAxes = self._viewMatrix

        self._blackWidget.hide()
        self._mainLayout.removeWidget(self._blackWidget)
        self._mainLayout.addWidget(self._vtkWidget)
        self._vtkWidget.show()

        # Start interaction
        self._renderWindow.GetInteractor().Start()

        # Trick to instantiate image property in iStyle
        self._iStyle.EndWindowLevel()
        self._iStyle.OnLeftButtonDown()
        self._iStyle.WindowLevel()
        self._renderWindow.GetInteractor().SetEventPosition(400, 0)
        self._iStyle.InvokeEvent(vtkCommand.StartWindowLevelEvent)
        self._iStyle.OnLeftButtonUp()
        self._iStyle.EndWindowLevel()

        if self._wwlEnabled:
            self._iStyle.StartWindowLevel()
            self._iStyle.OnLeftButtonUp()

        self._iStyle.SetCurrentImageNumber(0)

        self._renderer.ResetCamera()

        self._renderWindow.Render()

    def _setPrimaryName(self, name):
        self._textLayer.setPrimaryTextLine(2, name)

    @property
    def profileWidgetEnabled(self) -> bool:
        return self._profileWidget.enabled

    @profileWidgetEnabled.setter
    def profileWidgetEnabled(self, enabled: bool):
        if enabled==self._profileWidget.enabled:
            return

        if enabled and not self._profileWidget.enabled:
            self._profileWidgetNoInteractionYet = True
            self._profileWidget.callback = self._viewController.profileWidgetCallback
        else:
            self._profileWidgetNoInteractionYet = False
        self._profileWidget.enabled = enabled

        self.profileWidgeEnabledSignal.emit(self.profileWidgetEnabled)

    def setProfileWidgetEnabled(self, enabled):
        self.profileWidgetEnabled = enabled

    @property
    def secondaryImage(self) -> typing.Optional[Image2D]:
        if self._secondaryImageLayer.image is None:
            return None
        return self._secondaryImageLayer.image.data

    @secondaryImage.setter
    def secondaryImage(self, image: Image2D):
        if self.primaryImage is None:
            return

        imageAlreadyDisplayed = image == self._secondaryImageLayer.image or (not (self._secondaryImageLayer.image is None) and image == self._secondaryImageLayer.image.data)
        if imageAlreadyDisplayed:
            return

        self._secondaryImageLayer.image = Image2DForViewer(image)

        if image is None:
            self._secondaryImageLayer.image = None
            self.secondaryImageSignal.emit(self.secondaryImage)
            return

        self._secondaryImageLayer.resliceAxes = self._viewMatrix

        self._textLayer.setSecondaryTextLine(2, self.secondaryImage.name)

        if not (self._secondaryImageLayer.image is None):
            self._secondaryImageLayer.image.nameChangedSignal.disconnect(self._setSecondaryName)

        self._secondaryImageLayer.image.nameChangedSignal.connect(self._setSecondaryName)

        self._renderWindow.Render()

        self.secondaryImageSignal.emit(self.secondaryImage)

    def _setSecondaryName(self, name):
        self._textLayer.setSecondaryTextLine(2, name)

    @property
    def secondaryImageLayer(self):
        return self._secondaryImageLayer


    @property
    def viewType(self):
        return self._viewType

    @viewType.setter
    def viewType(self, viewType):
        if self._viewType == viewType:
            return

        self._setViewType(viewType)

    def _setViewType(self, viewType):
        self._viewType = viewType
        coronal = vtkCommonMath.vtkMatrix4x4()
        coronal.DeepCopy((1, 0, 0, 0,
                          0, 0, 1, 0,
                          0, 1, 0, 0,
                          0, 0, 0, 1))

        sagittal = vtkCommonMath.vtkMatrix4x4()
        sagittal.DeepCopy((0, 0, -1, 0,
                          1, 0, 0, 0,
                          0, 1, 0, 0,
                          0, 0, 0, 1))

        axial = vtkCommonMath.vtkMatrix4x4()
        axial.DeepCopy((1, 0, 0, 0,
                        0, -1, 0, 0,
                        0, 0, -1, 0,
                        0, 0, 0, 1))

        if self._viewType == self.ViewerTypes.SAGITTAL:
            self._viewMatrix = sagittal
        if self._viewType == self.ViewerTypes.AXIAL:
            self._viewMatrix = axial
        if self._viewType == self.ViewerTypes.CORONAL:
            self._viewMatrix = coronal
        else:
            ValueError('Invalid viewType')

        if not self.primaryImage is None:
            self._primaryImage2DLayer.resliceAxes = self._viewMatrix
            self._contourLayer.resliceAxes = self._viewMatrix
            self._rtPlanLayer.resliceAxes = self._viewMatrix
        if not self.secondaryImage is None:
            self._secondaryImageLayer.resliceAxes = self._viewMatrix

        self._renderer.ResetCamera()
        self._renderWindow.Render()

        self.viewTypeChangedSignal.emit(self._viewType)

    @property
    def crossHairEnabled(self) -> bool:
        return self._crossHairEnabled

    @crossHairEnabled.setter
    def crossHairEnabled(self, enabled: bool):
        if enabled == self._crossHairEnabled:
            return

        self._crossHairEnabled = enabled
        if self._crossHairEnabled:
            self.wwlEnabled = False
            self._crossHairLayer.visible = True
        else:
            self._crossHairLayer.visible = False
            self._handlePosition(None)
            self._renderWindow.Render()
        self.crossHairEnabledSignal.emit(self._crossHairEnabled)

    def setCrossHairEnabled(self, enabled: bool):
        self.crossHairEnabled = enabled

    @property
    def wwlEnabled(self) -> bool:
        return self._wwlEnabled

    @wwlEnabled.setter
    def wwlEnabled(self, enabled: bool):
        if enabled == self._wwlEnabled:
            return

        self._wwlEnabled = enabled

        if self._wwlEnabled:
            self.crossHairEnabled = False
        self.wwlEnabledSignal.emit(enabled)

    def setWWLEnabled(self, enabled: bool):
        self.wwlEnabled = enabled

    def onLeftButtonPressed(self, obj=None, event='Press'):
        if 'Press' in event:
            self._leftButtonPress = True

            if self.profileWidgetEnabled:
                self._iStyle.OnLeftButtonDown()
                return

            if self._crossHairEnabled:
                self.onMouseMove(None, None)
            else:
                self._iStyle.OnLeftButtonDown()
        else:
            self._leftButtonPress = False
            self._iStyle.OnLeftButtonUp()

    def onMouseMove(self, obj=None, event=None):
        (mouseX, mouseY) = self._renderWindow.GetInteractor().GetEventPosition()

        # MouseX and MouseY are not related to image but to renderWindow
        dPos = vtkCoordinate()
        dPos.SetCoordinateSystemToDisplay()
        dPos.SetValue(mouseX, mouseY, 0)
        worldPos = dPos.GetComputedWorldValue(self._renderer)

        point = self._viewMatrix.MultiplyPoint((worldPos[0], worldPos[1], 0, 1))

        if self.profileWidgetEnabled and self._profileWidgetNoInteractionYet and self._leftButtonPress:
            self._profileWidget.setInitialPosition((worldPos[0], worldPos[1]))
            self._profileWidgetNoInteractionYet = False
            return

        if self._crossHairEnabled and self._leftButtonPress:
            self._primaryImage2DLayer.image.selectedPosition = (point[0], point[1], point[2])

        if self._leftButtonPress and self._wwlEnabled:
            self._iStyle.OnMouseMove()
            self.__sendingWWL = True
            imageProperty = self._iStyle.GetCurrentImageProperty()
            self._primaryImage2DLayer.image.wwlValue = (imageProperty.GetColorWindow(), imageProperty.GetColorLevel())
            self.__sendingWWL = False

        if not self._leftButtonPress:
            self._iStyle.OnMouseMove()

    def onScroll(self, obj=None, event='Forward'):
        print(NotImplementedError)

    def _handlePosition(self, position: typing.Sequence):
        if not self._crossHairEnabled or position is None:
            self._textLayer.setPrimaryTextLine(0, '')
            self._textLayer.setPrimaryTextLine(1, '')

            if not self.secondaryImage is None:
                self._textLayer.setSecondaryTextLine(0, '')
                self._textLayer.setSecondaryTextLine(1, '')
            return

        transfo_mat = vtkCommonMath.vtkMatrix4x4()
        transfo_mat.DeepCopy(self._viewMatrix)
        transfo_mat.Invert()
        posAfterInverse = transfo_mat.MultiplyPoint((position[0], position[1], position[2], 1))

        pos = self._viewMatrix.MultiplyPoint((0, 0, posAfterInverse[2], 1))

        self._viewMatrix.SetElement(0, 3, pos[0])
        self._viewMatrix.SetElement(1, 3, pos[1])
        self._viewMatrix.SetElement(2, 3, pos[2])
        if self._crossHairEnabled:
            self._crossHairLayer.position = (posAfterInverse[0], posAfterInverse[1])

        try:
            data = self._primaryImage2DLayer.image.getDataAtPosition(position)

            self._textLayer.setPrimaryTextLine(0, 'Value: ' + "{:.2f}".format(data))
            # self._textLayer.setPrimaryTextLine(0, 'ValueNumpy: ' + "{:.2f}".format(dataNumpy))
            self._textLayer.setPrimaryTextLine(1,  'Pos: ' + "{:.2f}".format(position[0]) + ' ' + "{:.2f}".format(
                position[1]) + ' ' + "{:.2f}".format(position[2]))
        except:
            self._textLayer.setPrimaryTextLine(0, '')
            self._textLayer.setPrimaryTextLine(1, '')

        if not self.secondaryImage is None:
            try:
                data = self._secondaryImageLayer.image.getDataAtPosition(position)

                self._textLayer.setSecondaryTextLine(0, 'Value: ' + "{:.2f}".format(data))
                self._textLayer.setSecondaryTextLine(1, 'Pos: ' + "{:.2f}".format(position[0]) + ' ' + "{:.2f}".format(
                    position[1]) + ' ' + "{:.2f}".format(position[2]))
            except:
                self._textLayer.setSecondaryTextLine(0, '')
                self._textLayer.setSecondaryTextLine(1, '')



from PyQt5.QtCore import QTimer


class DynamicDisplayController():

    # MODE_DYNAMIC = 'DYNAMIC'
    # MODE_STATIC = 'STATIC'

    def __init__(self, viewController):

        self._viewController = viewController
        self._viewerPanelToolBar = None

        self.isDynamic = False
        self.dynamicViewerUnitList = []

        self.currentSpeedCoef = 1
        self.refreshRateInFramePerSec = 24
        self.refreshRateInMS = int(round(1000 / self.refreshRateInFramePerSec))
        self.timerStepNumber = 0
        self.time = 0
        self.timer = QTimer()
        self.timer.timeout.connect(self.checkIfUpdate)


    def setToolBar(self, viewerPanelToolBar):
        self._viewerPanelToolBar = viewerPanelToolBar
        self.connectToolBar()


    def connectToolBar(self):
        self._viewerPanelToolBar.playPauseSignal.connect(self.playOrPause)
        self._viewerPanelToolBar.fasterSignal.connect(self.playFaster)
        self._viewerPanelToolBar.slowerSignal.connect(self.playSlower)
        self._viewerPanelToolBar.refreshRateChangedSignal.connect(self.setRefreshRate)


    def addDynamicViewer(self, viewer):
        # print('in displayController addDynViewer', type(viewer))
        if viewer not in self.dynamicViewerUnitList:
            self.dynamicViewerUnitList.append(viewer)
        else:
            return
        if self.isDynamic == False:
            self.switchDynamicMode()


    def removeDynamicViewer(self, viewer):
        # print('in displayController removeDynamicViewer', type(viewer))
        if viewer in self.dynamicViewerUnitList:
            viewer.resetDynamicParameters()
            self.dynamicViewerUnitList.remove(viewer)
        else:
            return
        if not self.dynamicViewerUnitList:
            self.switchDynamicMode()


    def switchDynamicMode(self):
        """
        This function switches the mode from dynamic to static and inversely. It starts or stops the timer accordingly.
        """
        self.isDynamic = not self.isDynamic
        if self.isDynamic == True:
            print('Switch to dynamic mode')
            self.time = 0
            self.timer.start(self.refreshRateInMS)
            self._viewerPanelToolBar.addDynamicButtons()

        elif self.isDynamic == False:
            self.timer.stop()
            self._viewerPanelToolBar.removeDynamicButtons()
            print('Switch to static mode')


    def checkIfUpdate(self):
        """
        This function checks if an update must occur at this time.
        It only works for dynamic3DSequences for now.
        """
        self.time += self.refreshRateInMS * self.currentSpeedCoef
        for dynamicViewerUnit in self.dynamicViewerUnitList:

            # print(type(dynamicViewerUnit))

            dyn3DSeqForViewer = dynamicViewerUnit.primaryImage
            timingsList = dyn3DSeqForViewer.data.timingsList
            loopShift = timingsList[-1] * dynamicViewerUnit.loopStepNumber
            curIndex = dynamicViewerUnit.curPrimaryImgIdx

            # print('in dynamicDisplayController, checkIfUpdate', timingsList)

            if self.time - loopShift > timingsList[curIndex + 1]:
                newIndex = self.lookForClosestIndex(self.time - loopShift, curIndex, timingsList, dynamicViewerUnit)
                dynamicViewerUnit.nextImage(newIndex)


    def lookForClosestIndex(self, time, curIndex, timingsList, dataViewer):
        """
        This function return the index of the last position in timingList lower than time,
        meaning the time has passed this event and an update must occur.
        If the curIndex has reached the end of the timingsList, it returns 0
        """
        while timingsList[curIndex + 1] < time:
            curIndex += 1
            if curIndex == len(timingsList) - 1:  # has reach the end
                dataViewer.loopStepNumber += 1
                return 0

        return curIndex


    def playOrPause(self, playPauseBool):
        if playPauseBool:
            self.currentSpeedCoef = 1
        else:
            self.currentSpeedCoef = 0


    def playFaster(self):
        self.currentSpeedCoef *= 2


    def playSlower(self):
        self.currentSpeedCoef /= 2


    def setRefreshRate(self, refreshRate):
        self.refreshRateInFramePerSec = refreshRate
        if self.refreshRateInFramePerSec < 0.2:
            self.refreshRateInFramePerSec = 0.2
        if self.refreshRateInFramePerSec > 200:
            self.refreshRateInFramePerSec = 200
        self.refreshRateInMS = int(round(1000 / self.refreshRateInFramePerSec))
        self.timer.stop()
        self.timer.start(self.refreshRateInMS)
        print('Refresh Rate Set to', self.refreshRateInFramePerSec, 'frames/sec --> Check every', self.refreshRateInMS, 'ms')



import copy
from typing import Sequence, Optional

from PyQt5.QtCore import QDir
from PyQt5.QtWidgets import QLineEdit, QMenu, QAction, QInputDialog, QMessageBox, QMainWindow

from pydicom.uid import generate_uid

from opentps.core.data.images import CTImage
from opentps.core.data.images._image3D import Image3D
from opentps.core.data.dynamicData._dynamic3DSequence import Dynamic3DSequence
from opentps.core.data.dynamicData._dynamic3DModel import Dynamic3DModel
from opentps.core.data import PatientData
from opentps.core.io.serializedObjectIO import saveSerializedObjects
from opentps.gui.viewer.dataViewerComponents.patientDataPropertyEditor import PatientDataPropertyEditor


class PatientDataMenu:
    def __init__(self, viewController, parent=None):
        self._actions = []
        self._contextMenu = QMenu(parent)
        self._selectedData = None
        self._viewController = viewController

        self.dataPath = QDir.currentPath()  # maybe not the ideal default data directory

    @property
    def selectedData(self) -> Optional[Sequence[PatientData]]:
        return self._selectedData

    @selectedData.setter
    def selectedData(self, selectedData:Sequence[PatientData]):
        self._selectedData = selectedData

        self._buildMenu()

    def _buildMenu(self):
        self._actions = []

        dataClass = self._selectedData[0].__class__
        for data in self._selectedData:
            if data.__class__ != dataClass:
                dataClass = 'mixed'
                break

        if (len(self._selectedData) > 0):
            if len(self._selectedData)==1:
                self.rename_action = QAction("Rename")
                self.rename_action.triggered.connect(lambda checked: self._openRenameDataDialog(self._selectedData[0]))
                self._actions.append(self.rename_action)

                self.copy_action = QAction("Copy")
                self.copy_action.triggered.connect(lambda checked: self._copyData(self._selectedData[0]))
                self._actions.append(self.copy_action)

                self.info_action = QAction("Info")
                self.info_action.triggered.connect(lambda checked: self._showImageInfo(self._selectedData[0]))
                self._actions.append(self.info_action)

            if not dataClass == 'mixed':
                # actions for 3D images
                if (dataClass == Image3D or issubclass(dataClass, Image3D)) and len(self._selectedData) == 1:
                    self.superimpose_action = QAction("Superimpose")
                    self.superimpose_action.triggered.connect(lambda checked: self._setSecondaryImage(self._selectedData[0]))
                    self._actions.append(self.superimpose_action)

                # actions for group of 3DImage
                if (dataClass == CTImage or issubclass(dataClass, CTImage)) and len(self._selectedData) > 1:  # to generalize to other modalities eventually
                    self.make_series_action = QAction("Make dynamic 3D sequence")
                    self.make_series_action.triggered.connect(lambda checked: self._createDynamic3DSequence(self._selectedData))
                    self._actions.append(self.make_series_action)

            if dataClass == 'mixed':
                self.no_action = QAction("No action available for this group of data")
                self.context_menu.addAction(self.no_action)

            # actions for single Dynamic3DSequence
            if (dataClass == Dynamic3DSequence and len(self._selectedData) == 1):  # or dataClass == 'Dynamic2DSequence'):
                self.compute3DModelAction = QAction("Compute 4D model (MidP)")
                self.compute3DModelAction.triggered.connect(lambda checked, selected3DSequence=self._selectedData[0]: self._computeDynamic3DModel(selected3DSequence))
                self._actions.append(self.compute3DModelAction)

            self.delete_action = QAction("Delete")
            self.delete_action.triggered.connect(lambda checked: self._openDeleteDataDialog(self._selectedData))
            self._actions.append(self.delete_action)

            self.export_action = QAction("Export serialized")
            self.export_action.triggered.connect(lambda checked, selectedData=self._selectedData: self._exportSerializedData(selectedData))
            self._actions.append(self.export_action)

    def asContextMenu(self) -> QMenu:
        for action in self._actions:
            self._contextMenu.addAction(action)

        return self._contextMenu

    def _setSecondaryImage(self, image):
        self._viewController.secondaryImage = image

    def _copyData(self, selectedData):
        new_img = copy.deepcopy(selectedData)
        # new_img.patient = selectedData
        new_img.name = selectedData.name + '_copy'
        new_img.patient = selectedData.patient

    def _createDynamic3DSequence(self, selectedImages):
        newName, okPressed = QInputDialog.getText(None, "Set series name", "Series name:", QLineEdit.Normal, "4DCT")

        if (okPressed):
            Dynamic3DSequence.fromImagesInPatientList(selectedImages, newName)

    def _computeDynamic3DModel(self, selected3DSequence):
        newName, okPressed = QInputDialog.getText(None, "Set dynamic 3D model name", "3D model name:", QLineEdit.Normal, "MidP")

        if (okPressed):
            newMod = Dynamic3DModel()
            newMod.name = newName
            newMod.seriesInstanceUID = generate_uid()
            newMod.computeMidPositionImage(selected3DSequence)
            self._viewController.currentPatient.appendPatientData(newMod)

    def _openRenameDataDialog(self, data):
        text, ok = QInputDialog.getText(None, 'Rename data', 'New name:', text=data.name)
        if ok:
            data.name = str(text)

    def _openDeleteDataDialog(self, data):
        msgBox = QMessageBox()
        msgBox.setIcon(QMessageBox.Information)
        msgBox.setText("Delete data")
        msgBox.setWindowTitle("Delete selected data?")
        msgBox.setStandardButtons(QMessageBox.Ok | QMessageBox.Cancel)

        if msgBox.exec() == QMessageBox.Ok:
            for dataItem in data:
                dataItem.patient.removePatientData(dataItem)

    def _showImageInfo(self, image):
        w = QMainWindow(self._contextMenu.parent())
        w.setWindowTitle('Image info')
        w.resize(400, 400)
        w.setCentralWidget(PatientDataPropertyEditor(image, self._contextMenu.parent()))
        w.show()

    def _exportSerializedData(self, selectedData):
        from opentps.gui.panels.patientDataPanel.patientDataPanel import SaveData_dialog
        fileDialog = SaveData_dialog()
        savingPath, compressedBool, splitPatientsBool = fileDialog.getSaveFileName(None, dir=self.dataPath)

        saveSerializedObjects(selectedData, savingPath, compressedBool=compressedBool)



import os.path
from typing import Optional

from vtkmodules.vtkCommonTransforms import vtkTransform
from vtkmodules.vtkFiltersGeneral import vtkTransformPolyDataFilter
from vtkmodules.vtkIOGeometry import vtkSTLReader

from opentps.core.data.plan import RTPlan, PlanProtonBeam
from opentps.gui.viewer.dataForViewer.polyDataForViewer import PolyDataForViewer
from opentps.gui.viewer.dataViewerComponents.imageViewerComponents.polyData3DLayer_3D import PolyData3DLayer_3D
import opentps.gui.res.icons as iconModule

class BeamLayer_3D:
    def __init__(self, renderer, renderWindow):
        self._renderer = renderer
        self._renderWindow = renderWindow

        self._nozzleLayer = PolyData3DLayer_3D(self._renderer, self._renderWindow)

        filePath = os.path.join(iconModule.__path__[0], 'iba_nozzle.stl')
        self._stlReader = vtkSTLReader()
        self._stlReader.SetFileName(filePath)
        self._tformFilter = vtkTransformPolyDataFilter()
        self._tformFilter.SetTransform(self._tform(0, 0))
        self._tformFilter.SetInputConnection(self._stlReader.GetOutputPort())
        self._image = PolyDataForViewer(self._tformFilter)

    def close(self):
        self._nozzleLayer.close()
        self._tformFilter.RemoveAllInputs()

    def update(self):
        self._nozzleLayer.update()

    def _tform(self, gantryAngle, couchAngle):
        #TODO couchAngle
        tform = vtkTransform()
        tform.RotateY(-90)
        #tform.RotateY(90)
        tform.RotateZ(180)
        #tform.RotateX(90 + gantryAngle)
        tform.RotateX(180-gantryAngle)
        tform.Translate(0, 0, 0)

        return tform

    def setBeam(self, beam:PlanProtonBeam):
        self._tformFilter.SetTransform(self._tform(beam.gantryAngle, beam.couchAngle))
        self._nozzleLayer.image = self._image

class RTPlanLayer_3D:
    def __init__(self, renderer, renderWindow):
        self._renderer = renderer
        self._renderWindow = renderWindow

        self._beamLayers = []
        self._plan = None

    def close(self):
        for bLayer in self._beamLayers:
            bLayer.close()

        self._beamLayers = []

        self._renderWindow.Render()

    def update(self):
        for bl in self._beamLayers:
            bl.update()

    @property
    def plan(self) -> Optional[RTPlan]:
        return self._plan

    def setPlan(self, plan:RTPlan):
        #TODO connect to plan.dataChangedSignal
        if plan is None:
            self._plan = None
            self.close()
            return
        elif self._plan == plan:
            return

        self._plan = plan

        self.close()

        for beam in plan:
            bLayer = BeamLayer_3D(self._renderer, self._renderWindow)
            bLayer.setBeam(beam)
            self._beamLayers.append(bLayer)

        self._renderer.ResetCamera()
        self._renderWindow.Render()



import logging
from typing import Optional

from opentps.core.data.images import DoseImage, CTImage, MRImage
from opentps.core import Event
from opentps.gui.mainWindow import MainWindow
from opentps.gui.viewer.dataViewerComponents.profileWidget import ProfileWidget
from opentps.gui.viewer.dataViewer import DataViewer
from opentps.gui.viewer.dynamicDisplayController import DynamicDisplayController
from opentps.gui.viewer.viewerPanel import ViewerPanel


class ViewController():
    def __init__(self, patientList):
        # Events
        self.crossHairEnabledSignal = Event(bool)
        self.currentPatientChangedSignal = Event(object)
        self.independentViewsEnabledSignal = Event(bool)
        self.profileWidgetEnabledSignal = Event(object)
        self.mainImageChangedSignal = Event(object)
        self.dose1ChangedSignal = Event(object)
        self.dose2ChangedSignal = Event(object)
        self.patientAddedSignal = Event(object)
        self.patientRemovedSignal = Event(object)
        self.secondaryImageChangedSignal = Event(object)
        self.planChangedSignal = Event(object)
        self.showContourSignal = Event(object)
        self.windowLevelEnabledSignal = Event(bool)
        self.dropModeSignal = Event(object)
        self.droppedDataSignal = Event(object)
        self.displayLayoutTypeChangedSignal = Event(object)
        #self.dynamicViewerSwitchedOnSignal = Event(object)

        self.mainConfig = None

        self._activePatients = [patient for patient in patientList.patients]
        self._crossHairEnabled = None
        self._currentPatient = None
        self._dropMode = DataViewer.DropModes.DEFAULT
        self._droppedImage = None
        self._independentViewsEnabled = False
        self.profileWidgetCallback = ProfileWidget.ProfileWidgetCallback()
        self._profileWidgetEnabled = False
        self._mainImage = None
        self._dose1 = None
        self._dose2 = None
        self.multipleActivePatientsEnabled = False #TODO
        self._patientList = patientList
        self._plan = None
        self._selectedImage = None
        self._windowLevelEnabled = None
        self._displayLayout = ViewerPanel.LayoutTypes.DEFAULT
        self.shownDataUIDsList = []  # this is to keep track of which data is currently shown, but not used yet

        self.dynamicDisplayController = DynamicDisplayController(self)
        self.mainWindow = MainWindow(self)

        # self.dynamicDisplayController.connectViewerUnits(self.mainWindow.viewerPanel._viewerGrid)
        # self.dynamicDisplayController.setToolBar(self.mainWindow.viewerPanel._viewToolbar)

        self.logger = logging.getLogger(__name__)

        self._patientList.patientAddedSignal.connect(self._handleNewPatient)
        self._patientList.patientRemovedSignal.connect(self._handleRemovedPatient)

    def _handleNewPatient(self, patient):
        self._activePatients.append(patient)
        self.patientAddedSignal.emit(self._activePatients[-1])

        if self.currentPatient is None:
            self.currentPatient = patient

    def _displayCTOfCurrentPatient(self):
        if self.currentPatient is None:
            return
        
        ct = self.currentPatient.getPatientDataOfType(CTImage)
        if len(ct)>0:
            self.mainImage = ct[0]

    def _displayMRIOfCurrentPatient(self):
        if self.currentPatient is None:
            return

        mri = self.currentPatient.getPatientDataOfType(MRImage)
        if len(mri)>0:
            self.mainImage = mri[0]


    def _handleRemovedPatient(self, patient):
        self._activePatients.remove(patient)
        self.patientRemovedSignal.emit(patient)

        if self.currentPatient == patient:
            self.currentPatient = None

    @property
    def displayLayoutType(self):
        return self._displayLayout

    @displayLayoutType.setter
    def displayLayoutType(self, layout):
        if layout == self._displayLayout:
            return

        self._displayLayout = layout
        self.displayLayoutTypeChangedSignal.emit(self._displayLayout)

    @property
    def patientList(self):
        return self._patientList

    @property
    def activePatient(self):
        if self.multipleActivePatientsEnabled:
            self.logger.exception('Cannot getActivePatient if multiple patients enabled')
            raise

        if len(self._activePatients)>1:
            self.logger.exception('Multiple patients disabled but more than one active patient')
            raise

        return self._activePatients[0]

    @property
    def activePatients(self):
        return [patient for patient in self._activePatients]

    @property
    def crossHairEnabled(self):
        return self._crossHairEnabled

    @crossHairEnabled.setter
    def crossHairEnabled(self, enabled):
        if enabled==self._crossHairEnabled:
            return

        if self._windowLevelEnabled and enabled:
            self.windowLevelEnabled = False

        self._crossHairEnabled = enabled
        self.crossHairEnabledSignal.emit(self._crossHairEnabled)

    @property
    def currentPatient(self):
        return self._currentPatient

    @currentPatient.setter
    def currentPatient(self, patient):
        previousPatient = self._currentPatient
        noPreviousPatient = previousPatient is None

        if patient == previousPatient:
            return

        self._currentPatient = patient

        if noPreviousPatient:
            self._displayCTOfCurrentPatient()
            self._displayMRIOfCurrentPatient()
            self._currentPatient.patientDataAddedSignal.connect(self._handleNewCTinFirstPatient)
            self._currentPatient.patientDataAddedSignal.connect(self._handleNewMRIinFirstPatient)
        else:
            previousPatient.patientDataAddedSignal.disconnect(self._handleNewCTinFirstPatient)
            previousPatient.patientDataAddedSignal.disconnect(self._handleNewMRIinFirstPatient)

        self.currentPatientChangedSignal.emit(self._currentPatient)

    def _handleNewCTinFirstPatient(self, data):
        if isinstance(data, CTImage):
            self._displayCTOfCurrentPatient()
    
    def _handleNewMRIinFirstPatient(self, data):
        if isinstance(data, MRImage):
            self._displayMRIOfCurrentPatient()

    @property
    def dropMode(self):
        return self._dropMode

    @dropMode.setter
    def dropMode(self, mode):
        if mode == self._dropMode:
            return

        self._dropMode = mode
        self.dropModeSignal.emit(self._dropMode)

    @property
    def independentViewsEnabled(self):
        return self._independentViewsEnabled

    @independentViewsEnabled.setter
    def independentViewsEnabled(self, enabled):
        if enabled == self._independentViewsEnabled:
            return

        self._independentViewsEnabled = enabled

        self.independentViewsEnabledSignal.emit(self._independentViewsEnabled)

    @property
    def profileWidgetEnabled(self):
        return self._profileWidgetEnabled

    @profileWidgetEnabled.setter
    def profileWidgetEnabled(self, enabled):
        self._profileWidgetEnabled = enabled
        self.profileWidgetEnabledSignal.emit(self._profileWidgetEnabled)

    @property
    def mainImage(self):
        if self.independentViewsEnabled:
            # mainImage is only available when only one image can be shown
            raise Exception("mainImage is only available when only one image can be shown")
        return self._mainImage

    @mainImage.setter
    def mainImage(self, image):
        if self.independentViewsEnabled:
            # mainImage is only available when only one image can be shown
            raise Exception("mainImage is only available when only one image can be shown")

        self._mainImage = image
        self.mainImageChangedSignal.emit(self._mainImage)
        # self.dynamicOrStaticModeChangedSignal.emit(self._mainImage)
        if hasattr(self._mainImage, 'seriesInstanceUID') : 
            self.shownDataUIDsList.append(self._mainImage.seriesInstanceUID)

    @property
    def secondaryImage(self):
        if self.independentViewsEnabled:
            # secondaryImage is only available when only one image can be shown
            raise("secondaryImage is only available when only one image can be shown")
        return self._secondaryImage

    @secondaryImage.setter
    def secondaryImage(self, image):
        if self.independentViewsEnabled:
            # secondaryImage is only available when only one image can be shown
            raise Exception("secondaryImage is only available when only one image can be shown")

        self._secondaryImage = image
        self.secondaryImageChangedSignal.emit(self._secondaryImage)

    @property
    def dose1(self) -> Optional[DoseImage]:
        return self._dose1

    @dose1.setter
    def dose1(self, image:Optional[DoseImage]):
        if image == self._dose1:
            return

        self._dose1 = image
        self.dose1ChangedSignal.emit(self._dose1)

    @property
    def dose2(self) -> Optional[DoseImage]:
        return self._dose2

    @dose2.setter
    def dose2(self, image: Optional[DoseImage]):
        if image == self._dose2:
            return

        self._dose2 = image
        self.dose2ChangedSignal.emit(self._dose2)

    @property
    def plan(self):
        if self.independentViewsEnabled:
            # secondaryImage is only available when only one image can be shown
            raise ("plan is only available when only one image can be shown")
        return self._plan

    @plan.setter
    def plan(self, plan):
        if self.independentViewsEnabled:
            # secondaryImage is only available when only one image can be shown
            raise ("plan is only available when only one image can be shown")

        self._plan = plan
        self.planChangedSignal.emit(self._plan)

    @property
    def droppedImage(self):
        if self.independentViewsEnabled:
            # droppedImage is only available when only one image can be shown
            raise()

        return self._droppedImage

    @droppedImage.setter
    def droppedImage(self, image):
        if self.independentViewsEnabled:
            # droppedImage is only available when only one image can be shown
            raise()

        self._droppedImage = image
        self.droppedDataSignal.emit(self._droppedImage)

    @property
    def selectedImage(self):
        return self._selectedImage

    @selectedImage.setter
    def selectedImage(self, image):
        self._selectedImage = image

    @property
    def windowLevelEnabled(self):
        return self._windowLevelEnabled

    @windowLevelEnabled.setter
    def windowLevelEnabled(self, enabled):
        if enabled==self._windowLevelEnabled:
            return

        if self._crossHairEnabled and enabled:
            self.crossHairEnabled = False

        self._windowLevelEnabled = enabled
        self.windowLevelEnabledSignal.emit(self._windowLevelEnabled)

    def showContour(self, contour):
        self.showContourSignal.emit(contour)



from functools import partial
from typing import Union, Sequence, Optional

import numpy as np
from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import QApplication, QWidget, QVBoxLayout
from pyqtgraph import PlotWidget, mkPen, PlotCurveItem, SignalProxy, TextItem

from opentps.core.data.images import DoseImage
from opentps.core.data.images import ROIMask
from opentps.core.data import DVH
from opentps.core.data._roiContour import ROIContour
from opentps.core import Event
from opentps.gui.viewer.dataForViewer.ROIContourForViewer import ROIContourForViewer
from opentps.gui.viewer.dataForViewer.ROIMaskForViewer import ROIMaskForViewer


class DVHViewer(QWidget):
    def __init__(self, parent):
        super().__init__(parent=parent)

        self.doseChangeEvent = Event(object)
        self.dose2ChangeEvent = Event(object)

        self._dose = None
        self._rois = []
        self._dvhs = []
        self._dose2 = None
        self._dvhs2 = []
        self._partialVisibilityhandlers = []

        self._mainLayout = QVBoxLayout()
        self.setLayout(self._mainLayout)

        self._dvhPlot = DVHPlot(self)
        self._mainLayout.addWidget(self._dvhPlot)

    @property
    def dose(self) -> Optional[DoseImage]:
        return self._dose

    @dose.setter
    def dose(self, dose:DoseImage):
        if dose==self._dose:
            return

        self._dose = dose

        for dvh in self._dvhs:
            if self._dose is None:
                self._dvhPlot.removeDVH(dvh)
            else:
                dvh.dose = dose
                dvh.computeDVH()

        self.doseChangeEvent.emit(self._dose)

    @property
    def dose2(self) -> Optional[DoseImage]:
        return self._dose2

    @dose2.setter
    def dose2(self, dose: DoseImage):
        if dose == self._dose2:
            return

        self._dose2 = dose

        for dvh in self._dvhs2:
            if self._dose2 is None:
                self._dvhPlot.removeDVH(dvh)
            else:
                dvh.dose = dose
                dvh.computeDVH()

        self.dose2ChangeEvent.emit(self._dose2)

    @property
    def rois(self) -> Sequence[Union[ROIMask, ROIContour]]:
        return [roi for roi in self._rois]

    def appendROI(self, roi:Union[ROIMask, ROIContour]):
        if roi in self._rois:
            return

        # TODO a factory in dataForViewer would be nice because this small piece of code is often duplicated
        if isinstance(roi, ROIMask):
            roiForViewer = ROIMaskForViewer(roi)
        elif isinstance(roi, ROIContour):
            roiForViewer = ROIContourForViewer(roi)
        else:
            raise ValueError("ROI must be an instance of ROIMask or a ROIContour")

        if not roiForViewer.visible:
            return

        partialHandler = partial(self._handleROIVisibility, roi)
        roiForViewer.visibleChangedSignal.connect(partialHandler)

        dvh = DVH(roi)
        dvh2 = DVH(roi)

        self._rois.append(roi)
        self._partialVisibilityhandlers.append(partialHandler)

        self._dvhs.append(dvh)
        self._dvhPlot.appendDVH(dvh, roi, style=Qt.SolidLine)
        self._dvhs2.append(dvh2)
        self._dvhPlot.appendDVH(dvh2, roi, style=Qt.DashLine)

        if not (self._dose is None):
            dvh.dose = self._dose
            dvh.computeDVH()

        if not (self._dose2 is None):
            dvh2.dose = self._dose2
            dvh2.computeDVH()

    def _handleROIVisibility(self, roi, visibility):
        if not visibility:
            self.removeROI(roi)

    def removeROI(self, roi:Union[ROIMask, ROIContour]):
        partialHandler = self._partialVisibilityhandlers[self._rois.index(roi)]
        roiIndex = self._rois.index(roi)
        dvh = self._dvhs[roiIndex]

        # TODO a factory in dataForViewer would be nice because this small piece of code is often duplicated
        if isinstance(roi, ROIMask):
            roiForViewer = ROIMaskForViewer(roi)
        elif isinstance(roi, ROIContour):
            roiForViewer = ROIContourForViewer(roi)
        else:
            raise ValueError("ROI must be an instance of ROIMask or a ROIContour")

        roiForViewer.visibleChangedSignal.disconnect(partialHandler)

        self._dvhPlot.removeDVH(dvh)
        self._rois.remove(roi)

        dvh2Exists = len(self._dvhs2) > 0
        if dvh2Exists:
            dvh2 = self._dvhs2[roiIndex]
            self._dvhPlot.removeDVH(dvh2)
            self._dvhs2.remove(dvh2)
        self._dvhs.remove(dvh)

        self._partialVisibilityhandlers.remove(partialHandler)

    def clear(self):
        for roi in self._rois:
            self.removeROI(roi)
        self._dose = None
        self._dose2 = None


class DVHPlot(PlotWidget):
    def __init__(self, parent):
        PlotWidget.__init__(self, parent=parent)

        self.getPlotItem().setContentsMargins(5, 0, 20, 5)
        self.setBackground('k')
        self.setTitle("DVH")
        self.setLabel('left', 'Volume (%)')
        self.setLabel('bottom', 'Dose (Gy)')
        self.showGrid(x=True, y=True)
        self.setXRange(0, 100, padding=0)
        self.setYRange(0, 100, padding=0)

        self._parent = parent

        self._dvhs = []
        self._referenceROIs = []
        self._curves = []

    @property
    def DVHs(self) -> Sequence[DVH]:
        return [dvh for dvh in self._dvhs]

    def appendDVH(self, dvh:DVH, referenceROI:Union[ROIContour, ROIMask], style=Qt.SolidLine):
        curve = DVHCurve(dvh, referenceROI, self, style=style)
        self.addItem(curve.curve)
        self.addItem(curve.dvhLabel)

        self._referenceROIs.append(referenceROI)
        self._dvhs.append(dvh)
        self._curves.append(curve)

    def removeDVH(self, dvh:DVH):
        curve = self._curves[self._dvhs.index(dvh)]
        curve.clear()

        self._referenceROIs.remove(self._referenceROIs[self._dvhs.index(dvh)])
        self._curves.remove(curve)
        self._dvhs.remove(dvh)

    def clear(self):
        for curve in self._curves:
            curve.clear()

class DVHCurve:
    def __init__(self, dvh:DVH, referenceROI:Union[ROIContour, ROIMask], parent=None, style=Qt.SolidLine):
        self._dvh = dvh
        self._referenceROI = referenceROI
        self._parent = parent
        self._style = style

        self.curve = PlotCurveItem(np.array([]), np.array([]), style=style)

        self._dvh.dataUpdatedEvent.connect(self._setCurveData)
        self._referenceROI.nameChangedSignal.connect(self._setCurveData)
        self._referenceROI.colorChangedSignal.connect(self._setCurveData)

        self._viewer_DVH_proxy = SignalProxy(self._parent.scene().sigMouseMoved, rateLimit=120, slot=self._handleMouseMoved)
        self.dvhLabel = TextItem("", color=(255, 255, 255), fill=(0, 0, 0, 250), anchor=(0, 1))
        self.dvhLabel.hide()

        self._setCurveData()

    def _setCurveData(self, *args):
        mycolor = (self._referenceROI.color[0], self._referenceROI.color[1], self._referenceROI.color[2])
        pen = mkPen(color=mycolor, width=2, style=self._style)

        dose, volume = self._dvh.histogram
        self.curve.setData(dose, volume, pen=pen, name=self._referenceROI.name)
        self.curve.setData(dose, volume, pen=pen, name=self._referenceROI.name)
        # To force update the plot
        QApplication.processEvents()

    def clear(self):
        self.curve.setData(None, None)
        self._dvh.dataUpdatedEvent.disconnect(self._setCurveData)
        self._referenceROI.nameChangedSignal.disconnect(self._setCurveData)
        self._referenceROI.colorChangedSignal.disconnect(self._setCurveData)
        self.curve.clear() # TODO does nothing, apparently...

    def _handleMouseMoved(self, evt):
        self.dvhLabel.hide()

        mycolor = (self._referenceROI.color[0], self._referenceROI.color[1], self._referenceROI.color[2])

        if self.curve.sceneBoundingRect().contains(evt[0]):
            mousePoint = self._parent.getViewBox().mapSceneToView(evt[0])
            for item in self.curve.scene().items():
                if isinstance(item, PlotCurveItem):
                    data = item.getData()

                    if data[0] is None or len(data[0])<=0:
                        continue

                    y, y2 = np.interp([mousePoint.x(), mousePoint.x() * 1.01], data[0], data[1])
                    # if item.mouseShape().contains(mousePoint):
                    # check if mouse.y is close to f(mouse.x)
                    if abs(y - mousePoint.y()) < 2.0 + abs(y2 - y):  # abs(y2-y) is to increase the distance in high gradient
                        self.dvhLabel.setHtml("<b><font color='#" + "{:02x}{:02x}{:02x}".format(mycolor[0],
                                                                                                mycolor[1],
                                                                                                mycolor[2]) + "'>" + \
                                              item.name() + ":</font></b>" + \
                                                      "<br>D95 = {:.1f} Gy".format(self._dvh.D95) + \
                                                      "<br>D5 = {:.1f} Gy".format(self._dvh.D5) + \
                                                      "<br>Dmean = {:.1f} Gy".format(self._dvh.Dmean))
                        self.dvhLabel.setPos(mousePoint)
                        if (mousePoint.x() < 50 and mousePoint.y() < 50):
                            self.dvhLabel.setAnchor((0, 1))
                        elif (mousePoint.x() < 50 and mousePoint.y() >= 50):
                            self.dvhLabel.setAnchor((0, 0))
                        elif (mousePoint.x() >= 50 and mousePoint.y() < 50):
                            self.dvhLabel.setAnchor((1, 1))
                        elif (mousePoint.x() >= 50 and mousePoint.y() >= 50):
                            self.dvhLabel.setAnchor((1, 0))
                        self.dvhLabel.show()
                        break

class DVHBand:
    def __init__(self, dvh:DVH, referenceROI:Union[ROIContour, ROIMask], parent=None):
        self._dvh = dvh
        self._referenceROI = referenceROI
        self._parent = parent

        self.volume_low = []
        self.volume_high = []
        self.nominalDVH = []
        self.ROIDisplayColor = []
        self.Dmean = [0, 0]
        self.D98 = [0, 0]
        self.D95 = [0, 0]
        self.D50 = [0, 0]
        self.D5 = [0, 0]
        self.D2 = [0, 0]

    def compute_metrics(self):
        # compute metrics
        self.D98 = self.compute_band_Dx(98)
        self.D95 = self.compute_band_Dx(95)
        self.D50 = self.compute_band_Dx(50)
        self.D5 = self.compute_band_Dx(5)
        self.D2 = self.compute_band_Dx(2)

    def compute_band_Dx(self, x):
        dose = self._dvh.dose

        index = np.searchsorted(-self.volume_low, -x)
        if (index > len(self.volume_low) - 2): index = len(self.volume_low) - 2
        volume = self.volume_low[index]
        volume2 = self.volume_low[index + 1]
        if (volume == volume2):
            low_Dx = dose[index]
        else:
            w2 = (volume - x) / (volume - volume2)
            w1 = (x - volume2) / (volume - volume2)
            low_Dx = w1 * dose[index] + w2 * dose[index + 1]
            if low_Dx < 0: low_Dx = 0

        index = np.searchsorted(-self.volume_high, -x)
        if (index > len(self.volume_high) - 2): index = len(self.volume_high) - 2
        volume = self.volume_high[index]
        volume2 = self.volume_high[index + 1]
        if (volume == volume2):
            high_Dx = dose[index]
        else:
            w2 = (volume - x) / (volume - volume2)
            w1 = (x - volume2) / (volume - volume2)
            high_Dx = w1 * dose[index] + w2 * dose[index + 1]
            if high_Dx < 0: high_Dx = 0

        return [low_Dx, high_Dx]



import functools

from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import QVBoxLayout, QLabel, QLineEdit, QMainWindow, QWidget, QPushButton, QHBoxLayout, QCheckBox, \
    QFrame, QGroupBox

from opentps.core.processing.doseCalculation.doseCalculationConfig import DoseCalculationConfig
from opentps.gui.panels.mainToolbar import MainToolbar
from opentps.core.utils.programSettings import ProgramSettings


class EditableSetting(QWidget):
    def __init__(self, property, value, action, parent=None):
        super().__init__(parent)

        self._mainLayout = QHBoxLayout(self)
        self._mainLayout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(self._mainLayout)

        if not (property is None or property==""):
            self._txt = QLabel(self)
            self._txt.setText(property)
            self._mainLayout.addWidget(self._txt)

        self._nameLineEdit = QLineEdit(self)

        self._validateButton = QPushButton(self)
        self._validateButton.setText("Validate")
        self._validateButton.clicked.connect(lambda *args: action(self._nameLineEdit.text()))

        self.setValue(value)

        if not (property is None or property == ""):
            self._txt.setBuddy(self._nameLineEdit)

        self._mainLayout.addWidget(self._nameLineEdit)
        self._mainLayout.addWidget(self._validateButton)

    def setValue(self, value):
        self._nameLineEdit.setText(str(value))

class ProgramSettingEditor(QMainWindow):
    # singleton class!

    _staticVars = {"programSettings": None, "mainToolbar": None}

    def __init__(self):
        super().__init__()
        self.setWindowFlags(self.windowFlags() | Qt.WindowStaysOnTopHint)
        self.setWindowTitle('Program settings')
        self.resize(400, 400)

        centralWidget = QWidget()
        self.setCentralWidget(centralWidget)
        self._layout = QVBoxLayout()
        centralWidget.setLayout(self._layout)

        progSettingsTxt = 'Main program settings are located in ' + self.programSettings.programSettingsFolder
        self._progSettingsLabel = QLabel(self)
        self._progSettingsLabel.setText(progSettingsTxt)
        self._layout.addWidget(self._progSettingsLabel)

        self._workspaceField = EditableSetting("Workspace", str(self.programSettings.workspace), self.setWorkspace)
        self._layout.addWidget(self._workspaceField)

        self._doseCalculationFrame = QGroupBox(self)
        self._doseCalculationFrame.setTitle("Dose calculation settings")
        doseCalcLayout = QVBoxLayout(self._doseCalculationFrame)
        self._doseCalculationFrame.setLayout(doseCalcLayout)
        self._layout.addWidget(self._doseCalculationFrame)
        self._machineParam = MCsquareConfigEditor()
        doseCalcLayout.addWidget(self._machineParam)

        self._activeExtensions = ActiveExtensions(self.mainToolbar)
        self._layout.addWidget(self._activeExtensions)

    @property
    def programSettings(self) -> ProgramSettings:
        return self._staticVars["programSettings"]

    @staticmethod
    def setProgramSettings(config):
        ProgramSettingEditor._staticVars["programSettings"] = config

    @property
    def mainToolbar(self):
        return self._staticVars["mainToolbar"]

    @staticmethod
    def setMainToolbar(mainToolbar):
        ProgramSettingEditor._staticVars["mainToolbar"] = mainToolbar

    def setWorkspace(self, text):
        self.programSettings.workspace = text


class ActiveExtensions(QWidget):
    def __init__(self, toolbar:MainToolbar):
        super().__init__()

        self._mainLayout = QVBoxLayout(self)
        self.setLayout(self._mainLayout)

        for item in toolbar.items:
            itemCheckBox = QCheckBox(item.panelName)
            itemCheckBox.setChecked(item.visible)
            itemCheckBox.setCheckable(True)
            itemCheckBox.clicked.connect(functools.partial(self._handleItemChecked, item))

            self._mainLayout.addWidget(itemCheckBox)

    def _handleItemChecked(self, item, checked):
        item.visible = checked

class MCsquareConfigEditor(QWidget):
    def __init__(self, parent=None):
        super().__init__(parent)

        self._dcConfig = DoseCalculationConfig()

        self._layout = QVBoxLayout(self)
        self._layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(self._layout)

        self._txt = QLabel(self)
        self._txt.setText("Scanner folder:")
        self._layout.addWidget(self._txt)
        self._scannerField = EditableSetting("", str(self._dcConfig.scannerFolder), self._setScanner)
        self._layout.addWidget(self._scannerField)

        self._txt2 = QLabel(self)
        self._txt2.setText("BDL file:")
        self._layout.addWidget(self._txt2)
        self._bdlField = EditableSetting("", str(self._dcConfig.bdlFile), self._setBDL)
        self._layout.addWidget(self._bdlField)

        self._dcConfig.scannerFolderChangedSignal.connect(self._scannerField.setValue)
        self._dcConfig.bdlFileChangedSignal.connect(self._bdlField.setValue)

    def _setScanner(self, text):
        self._dcConfig.scannerFolder = text

    def _setBDL(self, text):
        self._dcConfig.bdlFile = text


from PyQt5.QtWidgets import QWidget, QVBoxLayout, QPushButton

class ExtensionPanel(QWidget):
  def __init__(self, viewController):
    QWidget.__init__(self)

    self._viewController = viewController

    self._layout = QVBoxLayout()
    self.setLayout(self._layout)

    self.flashTPSButton = QPushButton("This button does nothing")
    self._layout.addWidget(self.flashTPSButton)
